# -*- coding: utf-8 -*-


import numpy as np
import datastock as ds


from ..geom import CamLOS1D
from ..geom import _comp


# ##################################################################
# ##################################################################
#                   Rays
# ##################################################################


def _start_vect(ax=None, ay=None, az=None, name=None):
    dk = {
        f'x': ax,
        f'y': ay,
        f'z': az,
    }
    sh = None
    for k0, v0 in dk.items():
        dk[k0] = np.atleast_1d(v0).astype(float)
        if sh is None:
            sh = dk[k0].shape
        elif dk[k0].shape != sh:
            ls = [f"\t- {name}{k1}: {v1.shape}" for k1, v1 in dk.items()]
            msg = (
                f"Args {name}_x, {name}_y, {name}_z must have same shape!\n"
                "\n".join(ls)
            )
            raise Exception(msg)

    return dk['x'], dk['y'], dk['z']


def _check_inputs(
    coll=None,
    key=None,
    # start points
    start_x=None,
    start_y=None,
    start_z=None,
    # wavelength
    lamb=None,
    # ref
    ref=None,
    # from pts
    pts_x=None,
    pts_y=None,
    pts_z=None,
    # from ray-tracing (vect + length or config or diag)
    vect_x=None,
    vect_y=None,
    vect_z=None,
    length=None,
    config=None,
    reflections_nb=None,
    reflections_type=None,
    key_nseg=None,
    diag=None,
    key_cam=None,
):

    # -------------
    # key

    key = ds._generic_check._obj_key(
        d0=coll.dobj.get('rays', {}),
        short='ray',
        key=key,
    )

    # ------------
    # start points

    start_x, start_y, start_z = _start_vect(
        ax=start_x, ay=start_y, az=start_z, name='start',
    )
    shapes = [start_x.shape]

    # ------------
    # lamb

    if lamb is not None:
        lamb = np.atleast_1d(lamb).astype(float)

        if np.any(~np.isfinite(lamb)):
            msg = "Arg lamb must be all finite!"
            raise Exception(msg)

        if np.any(lamb <= 0.):
            msg = "ARg lamb must be all positive!"
            raise Exception(msg)

        shapes.append(lamb.shape)

    # ---------
    # shapes 1

    shref = None
    if len(shapes) == 1:
        if shapes[0] == (1,):
            pass
        else:
            shref = shapes[0]

    else:
        if shapes[0] == (1,) and shapes[1] == (1,):
            pass
        elif shapes[0] == shapes[1]:
            shref = shapes[0]
        elif (1,) in shapes:
            shref = shapes[1 - shapes.index((1,))]
        else:
            msg = (
                "Inconsistent shapes:\n"
                "\t- start_x: {shapes[0]}\n"
                "\t- lamb: {shapes[1]}\n"
            )
            raise Exception(msg)

    # ---------------------
    # pts vs config vs diag

    lc = [
        pts_x is not None,
        vect_x is not None
        and (
            config is not None
            or length is not None
            or diag is not None
        ),
    ]
    if np.sum(lc) != 1:
        msg = (
            "Please provide either:\n"
            "\t- pts_x, pts_y, pts_z: directly specify end points\n"
            "\t- config and / or diag: ray-tracing"
        )
        raise Exception(msg)

    # ------------
    # pts

    if lc[0]:

        pts_x, pts_y, pts_z = _start_vect(
            ax=pts_x, ay=pts_y, az=pts_z, name='pts',
        )

        if shref is None:
            if pts_x.ndim == 1:
                shref = pts_x.shape
                sh = tuple(np.r_[1, shref])
                pts_x = pts_x.reshape(sh)
                pts_y = pts_y.reshape(sh)
                pts_z = pts_z.reshape(sh)
            else:
                shref = pts_x.shape[1:]

        else:
            sh = tuple(np.r_[1, shref])
            if pts_x.shape == shref:
                pts_x = pts_x.reshape(sh)
                pts_y = pts_y.reshape(sh)
                pts_z = pts_z.reshape(sh)
            elif pts_x.ndim == len(shref) + 1 and pts_x.shape[1:] == shref:
                reflections_nb = pts_x.shape[0] - 1
            elif pts_x.shape == (1,):
                pts_x = np.full(shref, pts_x[0])
                pts_y = np.full(shref, pts_y[0])
                pts_z = np.full(shref, pts_z[0])
            else:
                msg = (
                    f"Arg pts_x has inconsistent shape:\n"
                    f"\t- pts_x.shape = {pts_x.shape}\n"
                    f"\t- shref: {shref}"
                )
                raise Exception(msg)

    # ------------
    # vectors

    elif lc[1]:

        vect_x, vect_y, vect_z = _start_vect(
            ax=vect_x, ay=vect_y, az=vect_z, name='vect',
        )

        # normalize
        norm = np.sqrt(vect_x**2 + vect_y**2 + vect_z**2)
        vect_x = vect_x / norm
        vect_y = vect_y / norm
        vect_z = vect_z / norm

        if shref is None:
            shref = vect_x.shape
        elif vect_x.shape not in [(1,), shref]:
            msg = (
                f"Arg vect_x has inconsistent shape {pts_x.shape} vs {shref}"
            )
            raise Exception(msg)

    # ----------
    # config

    # ---------
    # reflections

    # reflections_nb
    if reflections_nb is None:
        reflections_nb = 0

    reflections_nb = ds._generic_check._check_var(
        reflections_nb, 'reflections_nb',
        types=int,
        sign='>= 0',
    )

    # reflections_type
    reflections_type = ds._generic_check._check_var(
        reflections_type, 'reflections_type',
        types=str,
        default='specular',
        allowed=['specular', 'diffusive'],
    )

    if reflections_nb > 0 and config is None:
        msg = "reflections can only be handled if config is provided"
        raise Exception(msg)

    # key_nseg
    if key_nseg is not None:
        lok = list(coll.dref.keys())
        key_nseg = ds._generic_check._check_var(
            key_nseg, 'key_nseg',
            types=str,
            allowed=lok,
        )

    # ------
    # diag

    lspectro = None
    if diag is not None:
        lok = list(coll.dobj.get('diagnostic', {}).keys())
        diag = ds._generic_check._check_var(
            diag, 'diag',
            types=str,
            allowed=lok,
        )

        # key_cam
        lok = coll.dobj['diagnostic'][diag]['camera']
        key_cam = ds._generic_check._check_var(
            key_cam, 'key_cam',
            types=str,
            allowed=lok,
        )

        # lspectro
        lspectro = [
            oo for oo in coll.dobj['diagnostic'][diag]['doptics'][key_cam]['optics']
            if oo in coll.dobj.get('crystal', {}).keys()
            or oo in coll.dobj.get('grating', {}).keys()
        ]

    return (
        key,
        start_x, start_y, start_z,
        lamb, ref,
        pts_x, pts_y, pts_z,
        vect_x, vect_y, vect_z,
        shref,
        config, reflections_nb, reflections_type,
        key_nseg,
        diag, key_cam, lspectro,
    )


# ##################################################################
# ##################################################################
#                   Main
# ##################################################################


def _rays(
    coll=None,
    key=None,
    # start points
    start_x=None,
    start_y=None,
    start_z=None,
    # wavelength
    lamb=None,
    # ref
    ref=None,
    # from pts
    pts_x=None,
    pts_y=None,
    pts_z=None,
    # from ray-tracing (vect + length or config or diag)
    vect_x=None,
    vect_y=None,
    vect_z=None,
    length=None,
    config=None,
    reflections_nb=None,
    reflections_type=None,
    key_nseg=None,
    diag=None,
    key_cam=None,
):

    # -------------
    # check inputs

    (
        key,
        start_x, start_y, start_z,
        lamb, ref,
        pts_x, pts_y, pts_z,
        vect_x, vect_y, vect_z,
        shaperef,
        config, reflections_nb, reflections_type,
        key_nseg,
        diag, key_cam, lspectro,
    ) = _check_inputs(
        coll=coll,
        key=key,
        # start points
        start_x=start_x,
        start_y=start_y,
        start_z=start_z,
        # wavelength
        lamb=lamb,
        # ref
        ref=ref,
        # from pts
        pts_x=pts_x,
        pts_y=pts_y,
        pts_z=pts_z,
        # from ray-tracing (vect + length or config or diag)
        vect_x=vect_x,
        vect_y=vect_y,
        vect_z=vect_z,
        length=length,
        config=config,
        reflections_nb=reflections_nb,
        reflections_type=reflections_type,
        key_nseg=key_nseg,
        diag=diag,
        key_cam=key_cam,
    )

    # ----------------
    # prepare

    alpha, kalpha = None, None
    dalpha, kdalpha = None, None
    dbeta, kdbeta = None, None

    # -----------------------------
    # -----------------------------
    # compute from pts
    # ----------------------------

    if pts_x is not None:

        if pts_x.shape[0] == 1:
            angles = None
        else:
            v0x = pts_x[0, ...] - start_x
            v0y = pts_y[0, ...] - start_y
            v0z = pts_z[0, ...] - start_z
            vx = np.concatenate((v0x, np.diff(pts_x, axis=0)), axis=0)
            vy = np.concatenate((v0y, np.diff(pts_y, axis=0)), axis=0)
            vz = np.concatenate((v0z, np.diff(pts_z, axis=0)), axis=0)

            norm = np.sqrt(vx**2 + vy**2 + vz**2)
            vx = vx / norm
            vy = vy / norm
            vz = vz / norm

            sca = (
                vx[1:, ...] * vx[:-1, ...]
                + vy[1:, ...] * vy[:-1, ...]
                + vz[1:, ...] * vz[:-1, ...]
            )

            angle_in = None
            angle = np.arcos(sca)

    # -----------------------------
    # -----------------------------
    # compute ray-tracing
    # -----------------------------

    else:

        kk = None

        # final shape
        i0 = 0
        nbref = reflections_nb + 1
        if lspectro is not None and len(lspectro) > 0:
            i0 = len(lspectro)
            nbref += i0
        shape = tuple(np.r_[nbref, shaperef])

        # extract angles and pts
        pts_x = np.full(shape, np.nan)
        pts_y = np.full(shape, np.nan)
        pts_z = np.full(shape, np.nan)
        Rmin = np.full(shaperef, np.nan)

        pts_x[0, ...] = start_x
        pts_y[0, ...] = start_y
        pts_z[0, ...] = start_z

        if reflections_nb >= 0 or diag is not None:
            alpha = np.full(shape, np.nan)
            dalpha = np.full(shape, np.nan)
            dbeta = np.full(shape, np.nan)

        stx = np.copy(start_x)
        sty = np.copy(start_y)
        stz = np.copy(start_z)

        # ----------
        # diag

        if diag is not None:

            for ii, oo in enumerate(lspectro):

                reflect_ptsvect = coll.get_optics_reflect_ptsvect(oo)
                (
                    pts_x[ii, ...],
                    pts_y[ii, ...],
                    pts_z[ii, ...],
                    vect_x, vect_y, vect_z,
                    alpha[ii, ...],
                    iok,
                ) = reflect_ptsvect(
                    pts_x=stx,
                    pts_y=sty,
                    pts_z=stz,
                    vect_x=vect_x,
                    vect_y=vect_y,
                    vect_z=vect_z,
                    strict=True,
                    return_x01=False,
                )

                # update start
                stx[...] = pts_x[ii, ...]
                sty[...] = pts_y[ii, ...]
                stz[...] = pts_z[ii, ...]

        # ----------
        # config

        if config is not None:

            # prepare D
            D = np.array([
                stx.ravel(),
                sty.ravel(),
                stz.ravel(),
            ])

            # prepare u
            if vect_x.size == 1 and stx.size > 1:
                uu = np.array([
                    np.full(stx.shape, vect_x[0]),
                    np.full(stx.shape, vect_y[0]),
                    np.full(stx.shape, vect_z[0]),
                ])

            else:
                uu = np.array([
                    vect_x.ravel(),
                    vect_y.ravel(),
                    vect_z.ravel(),
                ])

            mask = np.all(np.isfinite(uu), axis=0)
            maskre = np.reshape(mask, shaperef)

            # call legacy code
            cam = CamLOS1D(
                dgeom=(D[:, mask], uu[:, mask]),
                config=config,
                Name='',
                Diag='',
                Exp='',
            )

            # add reflections
            if reflections_nb > 0:
                cam.add_reflections(nb=reflections_nb)

            if cam.dgeom['dreflect'] is not None:
                dref = cam.dgeom['dreflect']
                pts_x[i0:-1, maskre] = dref['Ds'][0, ...].T
                pts_y[i0:-1, maskre] = dref['Ds'][1, ...].T
                pts_z[i0:-1, maskre] = dref['Ds'][2, ...].T

                pout = (
                    dref['Ds'][:, :, -1]
                    + dref['kouts'][None, :, -1] * dref['us'][:, :, -1]
                )
            else:
                pout = cam.dgeom['PkOut']

            pts_x[-1, maskre] = pout[0, :]
            pts_y[-1, maskre] = pout[1, :]
            pts_z[-1, maskre] = pout[2, :]
            
            # RMin
            # kRMin = _comp.LOS_PRMin(cam.D, cam.u, kOut=None)
            # PRMin = cam.D + kRMin[None, :]*cam.u
            # Rmin[maskre] = np.hypot(PRMin[0, :], PRMin[1, :])

            vperp = cam.dgeom['vperp']
            u_perp = np.sum(cam.u*vperp, axis=0)

            alpha[i0, maskre] = np.arcsin(-u_perp)

            if cam.dgeom['dreflect'] is not None:
                us = dref['us'][..., 0]

                u_perp = np.sum(us * vperp, axis=0)
                dalpha[i0, maskre] = np.arcsin(u_perp) - alpha[0, maskre]

                # beta
                u0 = us - u_perp[None, ...] * vperp
                e0 = u0 / np.linalg.norm(u0)[None, ...]
                e1 = np.array([
                    vperp[1]*e0[2] - vperp[2]*e0[1],
                    vperp[2]*e0[0] - vperp[0]*e0[2],
                    vperp[0]*e0[1] - vperp[1]*e0[0],
                ])
                v0 = np.sum(us * e0, axis=0)
                v1 = np.sum(us * e1, axis=0)
                dbeta[i0, maskre] = np.arctan2(v1, v0)

        # ----------
        # length

        elif length is not None:

            shape = tuple(np.r_[1., shaperef])
            pts_x = (start_x + length * vect_x).reshape(shape)
            pts_y = (start_y + length * vect_y).reshape(shape)
            pts_z = (start_z + length * vect_z).reshape(shape)

    # --------------------------------
    # --------------------------------
    # store
    # --------------------------------

    # --------
    # dref

    shape = pts_x.shape
    nseg = shape[0]
    nextra = len(lspectro) if diag is not None else 0
    assert nseg == reflections_nb + 1 + nextra
    
    # key_nseg
    if key_nseg is None:
        knseg = f'{key}-nseg'
        dref = {
            knseg: {'size': nseg},
        }
    else:       
        if coll.dref[key_nseg]['size'] != nseg:
            msg = (
                "Wrong size of key_nseg:\n"
                f"\t- dref['key_nseg']['size] = coll.dref[key_nseg]['size']\n"
                f"\t- nseg = {nseg}"
            )
            raise Exception(msg)
            
        knseg = key_nseg
        dref = {}

    # if ref is None  
    if ref is None:
        ref = []
        for ii, ss in enumerate(shaperef):
            kr = f'{key}-n{ii}'
            dref.update({
                kr: {'size': ss}
            })
            ref.append(kr)
        ref = tuple(ref)

    refpts = tuple(np.r_[(knseg,), ref])

    #  --------
    # ddata

    # pts
    kpx = f'{key}_ptx'
    kpy = f'{key}_pty'
    kpz = f'{key}_ptz'
    # kRmin = f'{key}-Rmin'

    ddata = {
        kpx: {
            'data': pts_x,
            'ref': refpts,
            'dim': 'distance',
            'quant': 'distance',
            'name': 'x',
            'units': 'm',
        },
        kpy: {
            'data': pts_y,
            'ref': refpts,
            'dim': 'distance',
            'quant': 'distance',
            'name': 'y',
            'units': 'm',
        },
        kpz: {
            'data': pts_z,
            'ref': refpts,
            'dim': 'distance',
            'quant': 'distance',
            'name': 'z',
            'units': 'm',
        },
        # kRmin: {
        #     'data': Rmin,
        #     'ref': refpts[1:],
        #     'dim': 'distance',
        #     'quant': 'distance',
        #     'name': 'R',
        #     'units': 'm',
        # },
    }

    # start
    if start_x.shape != (1,):
        ksx = f'{key}_startx'
        ksy = f'{key}_starty'
        ksz = f'{key}_startz'
        if ref is None:
            ref = None

        ddata.update({
            ksx: {
                'data': start_x,
                'ref': ref,
                'dim': 'distance',
                'quant': 'distance',
                'name': 'x',
                'units': 'm',
            },
            ksy: {
                'data': start_y,
                'ref': ref,
                'dim': 'distance',
                'quant': 'distance',
                'name': 'y',
                'units': 'm',
            },
            ksz: {
                'data': start_z,
                'ref': ref,
                'dim': 'distance',
                'quant': 'distance',
                'name': 'z',
                'units': 'm',
            },
        })

    # alpha
    if alpha is not None:
        kalpha = f'{key}_alpha'
        ddata.update({
            kalpha: {
                'data': alpha,
                'ref': refpts,
                'dim': 'angle',
                'quant': 'angle',
                'name': 'alpha',
                'units': 'rad',
            },
        })

    # dalpha
    if dalpha is not None:
        kdalpha = f'{key}_dalpha'
        ddata.update({
            kdalpha: {
                'data': dalpha,
                'ref': refpts,
                'dim': 'angle',
                'quant': 'angle',
                'name': 'alpha',
                'units': 'rad',
            },
        })

    # dbeta
    if dbeta is not None:
        kdbeta = f'{key}_dbeta'
        ddata.update({
            kdbeta: {
                'data': dbeta,
                'ref': refpts,
                'dim': 'angle',
                'quant': 'angle',
                'name': 'beta',
                'units': 'rad',
            },
        })

    # --------
    # dobj

    # start
    if start_x.shape == (1,):
        start = np.r_[start_x, start_y, start_z]
    else:
        start = (ksx, ksy, ksz)

    # lamb
    if lamb is not None:
        if lamb.shape == (1,):
            ll = lamb
        else:
            ll = lamb
    else:
        ll = lamb

    # pts
    pts = (kpx, kpy, kpz)

    # dobj
    dobj = {
        'rays': {
            key: {
                'start': start,
                'pts': pts,
                'lamb': ll,
                'shape': shape,
                'ref': refpts,
                # 'Rmin': kRmin,
                'alpha': kalpha,
                'reflect_dalpha': kdalpha,
                'reflect_dbeta': kdbeta,
            },
        },
    }

    return dref, ddata, dobj

# ##################################################################
# ##################################################################
#                   Rays - check key
# ##################################################################


def _check_key(coll=None, key=None, key_cam=None):

    # check key
    lrays = list(coll.dobj.get('rays', {}).keys())
    ldiag = [
        k0 for k0, v0 in coll.dobj.get('diagnostic', {}).items()
        if any([
                v1.get('los') is not None
                and v1['los'] in lrays
                for k1, v1 in v0['doptics'].items()
            ])
    ]
    
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lrays + ldiag,
    )
    
    # Derive kray
    if key in lrays:
        kray = key
    else:
        
        # key_cam
        lok = list(coll.dobj['diagnostic'][key]['doptics'].keys())
        key_cam = ds._generic_check._check_var(
            key_cam, 'key_cam',
            types=str,
            allowed=lok,
        )    
        
        kray = coll.dobj['diagnostic'][key]['doptics'][key_cam]['los']
    
    return kray


# ##################################################################
# ##################################################################
#                   Rays - get start and vect
# ##################################################################


def _get_start(
    coll=None,
    key=None,
    key_cam=None,
):

    # ---------
    # check key

    key = _check_key(coll=coll, key=key, key_cam=key_cam)

    # ---------------
    # get start

    stx, sty, stz = coll.dobj['rays'][key]['start']
    if isinstance(stx, str):
        stx = coll.ddata[stx]['data']
        sty = coll.ddata[sty]['data']
        stz = coll.ddata[stz]['data']

    return stx, sty, stz


def _get_pts(
    coll=None,
    key=None,
    key_cam=None,
):

    # ---------
    # check key

    key = _check_key(coll=coll, key=key, key_cam=key_cam)
    
    # ---------
    # get start

    stx, sty, stz = _get_start(coll=coll, key=key)

    # ---------------
    # get other pts

    ptsx, ptsy, ptsz = coll.dobj['rays'][key]['pts']
    ptsx = coll.ddata[ptsx]['data']
    ptsy = coll.ddata[ptsy]['data']
    ptsz = coll.ddata[ptsz]['data']

    # concatenate
    ptsx = np.insert(ptsx, 0, stx, axis=0)
    ptsy = np.insert(ptsy, 0, sty, axis=0)
    ptsz = np.insert(ptsz, 0, stz, axis=0)

    return ptsx, ptsy, ptsz


def _get_vect(
    coll=None,
    key=None,
    key_cam=None,
    norm=None,
):

    # ---------
    # check key

    key = _check_key(coll=coll, key=key, key_cam=key_cam)

    # norm
    norm = ds._generic_check._check_var(
        norm, 'norm',
        types=bool,
        default=True,
    )

    # ---------------
    # get start

    stx, sty, stz = _get_start(coll=coll, key=key)

    ptsx, ptsy, ptsz = coll.dobj['rays'][key]['pts']
    ptsx = coll.ddata[ptsx]['data']
    ptsy = coll.ddata[ptsy]['data']
    ptsz = coll.ddata[ptsz]['data']

    vx = (ptsx[0, ...] - stx)[None, ...]
    vy = (ptsy[0, ...] - sty)[None, ...]
    vz = (ptsz[0, ...] - stz)[None, ...]
    if ptsx.shape[0] > 1:
        vx = np.concatenate((vx, np.diff(ptsx, axis=0)), axis=0)
        vy = np.concatenate((vy, np.diff(ptsy, axis=0)), axis=0)
        vz = np.concatenate((vz, np.diff(ptsz, axis=0)), axis=0)

    # normalize
    if norm is True:
        norm = np.sqrt(vx**2 + vy**2 + vz**2)
        vx = vx / norm
        vy = vy / norm
        vz = vz / norm

    return vx, vy, vz


# ##################################################################
# ##################################################################
#                   Rays - remove
# ##################################################################


def _remove(
    coll=None,
    key=None,
):

    # ----------
    # check

    lok = list(coll.dobj.get('rays', {}).keys())
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lok,
    )

    # ------------
    # list data

    lkd = [
        'start', 'pts', 'lamb',
        'alpha', 'reflect_dalpha', 'reflect_dbeta',
    ]
    ld = []
    for k0 in lkd:
        v0 = coll._dobj['rays'][key][k0]
        if isinstance(v0, str):
            ld.append(v0)
        elif isinstance(v0, (tuple, list)):
            ld += list(v0)
        elif v0 is None:
            pass
        else:
            msg = "How to handle rays['{key}']['{k0}'] = {v0}"
            raise Exception(msg)

    # -----------
    # remove data

    coll.remove_data(ld, propagate=True)

    # -----------
    # remove rays

    del coll._dobj['rays'][key]




# Built-in
import os
import warnings

# Common
import scipy.io as scpio
import numpy as np

# tofu-specific
from .. import _physics


__all__ = ['get_data_from_matids']
_LIDSOK = ['core_profiles']
_DRETURN = {'core_profiles': ['rhotn', 'ne', 'Te', 'zeff', 't', 'brem']}
_MSG0 = ("The input file structure is not as expected !\n"
         + "  => Maybe file structure changed ?\n"
         + "  => Maybe corrupted data ?\n")


# ####################################################
#               Utility
# ####################################################

def _get_indtlim(t, tlim=None, shot=None, out=bool):
    """  """
    c0 = tlim is None
    c1 = type(tlim) in [list, tuple, np.ndarray]
    assert c0 or c1
    assert type(t) is np.ndarray

    if c0:
        tlim = [-np.inf, np.inf]
    else:
        assert len(tlim) == 2
        ls = [int, float, np.int64, np.float64]     # , str
        assert all([tt is None or type(tt) in ls for tt in tlim])
        tlim = list(tlim)
        for (ii, sgn) in [(0, -1.), (1, 1.)]:
            if tlim[ii] is None:
                tlim[ii] = sgn*np.inf
            # elif type(tlim[ii]) is str and 'ign' in tlim[ii].lower():
                # tlim[ii] = get_t0(shot)

    assert tlim[0] < tlim[1]
    indt = (t >= tlim[0]) & (t <= tlim[1])
    if out is int:
        indt = indt.nonzero()[0]
    return indt


# ####################################################
#               Main function
# ####################################################

def get_data_from_matids(input_pfe=None, tlim=None,
                         return_fields=None, lamb=None):
    """ Extract tofu-compatible from an ids saved as a mat file

    Assumes that the mat file contains the ids data
    Only the following ids are handled:
        {}

    """.format(_LIDSOK)

    # ---------------
    # Check
    if not os.path.isfile(input_pfe):
        msg = ("Provided file does not seem to exist:\n"
               + "\t- {}".format(input_pfe))
        raise Exception(msg)
    lc = [return_fields is None,
          isinstance(return_fields, str),
          isinstance(return_fields, list)
          and all([isinstance(ss, str) for ss in return_fields])]
    if not any(lc):
        msg = "return_fields must be a str or a list of str "
        raise Exception(msg)
    if lc[1]:
        return_fields = [return_fields]

    # ---------------
    # Load and check / extract ids
    mat = scpio.loadmat(input_pfe)
    ids = [k0 for k0 in mat.keys() if '__' not in k0]
    if len(ids) != 1 or ids[0] not in _LIDSOK:
        msg = ("The file does not seem to contain a known ids:\n"
               + "\t- file: {}\n".format(input_pfe)
               + "\t- keys: {}\n".format(sorted(mat.keys()))
               + "\t- known ids: {}".format(_LIDSOK))
        raise Exception(msg)
    ids = ids[0]
    data = mat[ids]

    if return_fields is None:
        return_fields = _DRETURN[ids]
    notok = [ss for ss in return_fields if ss not in _DRETURN[ids]]
    if len(notok) > 0:
        msg = ("Some requested fields are not available:\n"
               + "\t- requested: {}\n".format(notok)
               + "\t- available: {}".format(_DRETURN[ids]))
        raise Exception(msg)

    # ---------------
    # Get inside ids and extract data
    if ids == 'core_profiles':
        # ---------------
        # Check expected structure
        if not (data.shape == (1, 1) and data[0, 0].size == 1):
            msg = ("\t{}.shape = {}\n".format(ids, data.shape)
                   + "\t{}.size = {}".format(ids, data[0, 0].size))
            raise Exception(_MSG0 + msg)
        data = data[0, 0].tolist()
        if not (isinstance(data, tuple) and len(data) == 6):
            msg = ("\ttype({}[0, 0].tolist()) = {}\n".format(ids, type(data))
                   + "\tlen({}[0, 0].tolist()) = {}".format(ids, len(data)))
            raise Exception(_MSG0 + msg)

        ls = [pp.shape for pp in data]
        c0 = [len(ss) == 2 for ss in ls]
        c1 = np.sum([ss == (1, 1) for ss in ls]) == 4
        c2 = np.sum([(ss[1] == 1 and ss[0] > ss[1]) for ss in ls]) == 1
        c3 = np.sum([(ss[0] == 1 and ss[1] > ss[0]) for ss in ls]) == 1
        if c0 and c1 and c2 and c3:
            indt = [ii for ii in range(len(ls)) if ls[ii][0] > ls[ii][1]]
            indp = [ii for ii in range(len(ls)) if ls[ii][0] < ls[ii][1]]
        else:
            if np.sum([ss == (1, 1) for ss in ls]) == 6:
                warnings.warn("There seems to be only one time step...")
                indt = [ii for ii in range(len(ls)) if data[ii].dtype == '<f8']
                indp = [ii for ii in range(len(ls)) if data[ii].dtype == 'O']
            else:
                msg = "\t{} contains shapes {}".format(ids, ls)
                raise Exception(_MSG0 + msg)

        if len(indt) != 1 or len(indp) != 1:
            msg = ("\tseveral options for time / profile arrays:\n"
                   + "\t\t- len(indt) = {}\n".format(len(indt))
                   + "\t\t- len(indp) = {}".format(len(indp)))
            raise Exception(_MSG0 + msg)
        indt, indp = indt[0], indp[0]

        if not data[indt].size == data[indp].size:
            msg = "\tTime vector and profiles have different sizes !"
            raise Exception(_MSG0 + msg)

        # ---------------
        # Get time vector
        t = data[indt].ravel().astype(float)
        indt = _get_indtlim(t, tlim=tlim, shot=None, out=int)
        t = t[indt]
        nt = t.size

        dout = {}
        if 't' in return_fields:
            dout['t'] = t

        # ---------------
        # Continue checks and get indices of quantities
        data = data[indp].ravel()[indt]
        assert all([pp.shape == (1, 1) for pp in data])

        des = [ss[0] for ss in data[0][0, 0].dtype.descr]

        if 'rhotn' in return_fields:
            indg = [ii for ii in range(len(des)) if des[ii] == 'grid'][0]
            desg = [ss[0] for ss in data[0][0, 0][indg].dtype.descr]
            indrhotn = [ii for ii in range(len(desg))
                        if desg[ii] == 'rho_tor_norm'][0]
            dout['rhotn'] = np.array([
                data[ii][0, 0][indg][0, 0][indrhotn].ravel()
                for ii in range(nt)])

        if 'brem' in return_fields or 'zeff' in return_fields:
            indZeff = [ii for ii in range(len(des))
                       if des[ii] == 'zeff'][0]
            zeff = np.array([data[ii][0, 0][indZeff].ravel()
                             for ii in range(nt)])
            if 'zeff' in return_fields:
                dout['zeff'] = zeff

        if any([ss in return_fields for ss in ['brem', 'Te', 'ne']]):
            inde = [ii for ii in range(len(des)) if des[ii] == 'electrons'][0]
            dese = [ss[0] for ss in data[0][0, 0][inde].dtype.descr]

        if 'brem' in return_fields or 'Te' in return_fields:
            indTe = [ii for ii in range(len(dese))
                     if dese[ii] == 'temperature'][0]
            Te = np.array([data[ii][0, 0][inde][0, 0][indTe].ravel()
                           for ii in range(nt)])
            if 'Te' in return_fields:
                dout['Te'] = Te

        if 'brem' in return_fields or 'ne' in return_fields:
            indne = [ii for ii in range(len(dese))
                     if dese[ii] == 'density'][0]
            ne = np.array([data[ii][0, 0][inde][0, 0][indne].ravel()
                           for ii in range(nt)])
            if 'ne' in return_fields:
                dout['ne'] = ne

        if 'brem' in return_fields:
            dout['brem'] = _physics.compute_bremzeff(Te=Te, ne=ne,
                                                     zeff=zeff, lamb=lamb)[0]
    return dout



# -*- coding: utf-8 -*-


import numpy as np
import datastock as ds


# ##############################################################
# ##############################################################
#           Finding reflection points
# ##############################################################


def _get_ptsvect(
    coll=None,
    key=None,
    asplane=None,
):

    # ---------
    # key

    key, cls = coll.get_optics_cls(optics=key)
    key, cls = key[0], cls[0]
    dgeom = coll.dobj[cls][key]['dgeom']

    asplane = ds._generic_check._check_var(
        asplane, 'asplane',
        types=bool,
        default=False,
    )

    # -------------------
    #     Planar
    # -------------------

    if dgeom['type'] == 'planar' or asplane is True:

        if dgeom['extenthalf'] is None:
            x0max, x1max = None, None
        else:
            x0max, x1max = dgeom['extenthalf']

        ptsvect = _get_ptsvect_plane(
            plane_cent=dgeom['cent'],
            plane_nin=dgeom['nin'],
            plane_e0=dgeom['e0'],
            plane_e1=dgeom['e1'],
            # limits
            x0max=x0max,
            x1max=x1max,
        )

    # ----------------
    #   Cylindrical
    # ----------------

    elif dgeom['type'] == 'cylindrical':

        iplan = np.isinf(dgeom['curve_r']).nonzero()[0][0]
        icurv = 1 - iplan
        eax = ['e0', 'e1'][iplan]
        erot = ['e0', 'e1'][icurv]
        
        rc = dgeom['curve_r'][icurv]
        rcs = np.sign(rc)
        rca = np.abs(rc)
        
        minmax = np.maximum if rcs > 0 else np.minimum

        def ptsvect(
            # pts
            pts_x=None,
            pts_y=None,
            pts_z=None,
            # vect
            vect_x=None,
            vect_y=None,
            vect_z=None,
            # surface
            O=dgeom['cent'] + dgeom['nin'] * rc,
            rcs=rcs,
            rca=rca,
            minmax=minmax,
            eax=dgeom[eax],
            erot=dgeom[erot],
            iplan=iplan,
            # limits
            thetamax=dgeom['extenthalf'][icurv],
            xmax=dgeom['extenthalf'][iplan],
            # local coordinates
            nin=dgeom['nin'],
            # return
            strict=None,
            return_x01=None,
        ):
            """

            cf. Notes_Upgrades/raytracing_surface3draytracing_surface3d.pdf

            """
            # normalize vect
            vn = np.sqrt(vect_x**2 + vect_y**2 + vect_z**2)
            vect_x = vect_x / vn
            vect_y = vect_y / vn
            vect_z = vect_z / vn

            # ------------------------------------------
            # Get local coordinates of reflection points

            # get parameters for k
            OAzx = (pts_y - O[1])*eax[2] - (pts_z - O[2])*eax[1]
            OAzy = (pts_z - O[2])*eax[0] - (pts_x - O[0])*eax[2]
            OAzz = (pts_x - O[0])*eax[1] - (pts_y - O[1])*eax[0]
            OAz2 = OAzx**2 + OAzy**2 + OAzz**2

            eazx = vect_y*eax[2] - vect_z*eax[1]
            eazy = vect_z*eax[0] - vect_x*eax[2]
            eazz = vect_x*eax[1] - vect_y*eax[0]
            ez2 = eazx**2 + eazy**2 + eazz**2

            OAzez = OAzx*eazx + OAzy*eazy + OAzz*eazz
            iok = (
                np.isfinite(ez2)
                & np.isfinite(OAz2)
                & np.isfinite(OAzez)
            )

            # Prepare D, theta, xx
            shape = OAzez.shape
            (
                Dx, Dy, Dz, vrx, vry, vrz,
                angle, theta, xx, kk,
            ) = _common_prepare(shape)

            # get k
            delta = (OAzez**2 - ez2*(OAz2-rca**2))[iok]
            ipos = delta >= 0
            iok[iok] = ipos
            sol0 = (-OAzez[iok] - np.sqrt(delta[ipos])) / ez2[iok]
            sol1 = (-OAzez[iok] + np.sqrt(delta[ipos])) / ez2[iok]
            kk[iok] = minmax(sol0, sol1)

            if np.any(iok):
                Dx = pts_x + kk*vect_x
                Dy = pts_y + kk*vect_y
                Dz = pts_z + kk*vect_z

                ODx = Dx[iok] - O[0]
                ODy = Dy[iok] - O[1]
                ODz = Dz[iok] - O[2]

                xxi = ODx * eax[0] + ODy * eax[1] + ODz * eax[2]
                
                nox = ODx - xxi * eax[0]
                noy = ODy - xxi * eax[1]
                noz = ODz - xxi * eax[2]
                nn = np.sqrt(nox**2 + noy**2 + noz**2)
                nox = rcs * nox / nn
                noy = rcs * noy / nn
                noz = rcs * noz / nn
                

                # ODzx = (Dy[iok] - O[1])*eax[2] - (Dz[iok] - O[2])*eax[1]
                # ODzy = (Dz[iok] - O[2])*eax[0] - (Dx[iok] - O[0])*eax[2]
                # ODzz = (Dx[iok] - O[0])*eax[1] - (Dy[iok] - O[1])*eax[0]
                # ODzn = np.sqrt(ODzx**2 + ODzy**2 + ODzz**2)

                # nox = -rcs*(ODzy * eax[2] - ODzz * eax[1]) / ODzn
                # noy = -rcs*(ODzz * eax[0] - ODzx * eax[2]) / ODzn
                # noz = -rcs*(ODzx * eax[1] - ODzy * eax[0]) / ODzn

                # scalar product (for angle + reflection)
                if np.isscalar(vect_x):
                    scavn = -(vect_x*nox + vect_y*noy + vect_z*noz)

                    # get vect_reflect
                    vrx[iok] = vect_x + 2.*scavn * nox
                    vry[iok] = vect_y + 2.*scavn * noy
                    vrz[iok] = vect_z + 2.*scavn * noz

                else:
                    scavn = -(
                        vect_x[iok]*nox + vect_y[iok]*noy + vect_z[iok]*noz
                    )
                    # get vect_reflect
                    vrx[iok] = vect_x[iok] + 2.*scavn * nox
                    vry[iok] = vect_y[iok] + 2.*scavn * noy
                    vrz[iok] = vect_z[iok] + 2.*scavn * noz

                angle[iok] = -np.arcsin(scavn)

                # x0, x1
                if strict is True or return_x01 is True:
                    
                    theta[iok] =  rcs * np.arctan2(
                        nox*erot[0] + noy*erot[1] + noz*erot[2],
                        -nox*nin[0] - noy*nin[1] - noz*nin[2],
                    )

                    xx[iok] = xxi

                    if strict is True:
                        iout = (np.abs(theta) > thetamax) | (np.abs(xx) > xmax)

                        if np.any(iout):
                            Dx[iout] = np.nan
                            Dy[iout] = np.nan
                            Dz[iout] = np.nan
                            vrx[iout] = np.nan
                            vry[iout] = np.nan
                            vrz[iout] = np.nan
                            angle[iout] = np.nan
                            iok[iout] = False

                # enforce normalization
                vnorm = np.sqrt(vrx[iok]**2 + vry[iok]**2 + vrz[iok]**2)
                vrx[iok] = vrx[iok] / vnorm
                vry[iok] = vry[iok] / vnorm
                vrz[iok] = vrz[iok] / vnorm

            # return
            if return_x01:
                if iplan == 0:
                    return Dx, Dy, Dz, vrx, vry, vrz, angle, iok, xx, theta
                else:
                    return Dx, Dy, Dz, vrx, vry, vrz, angle, iok, theta, xx
            else:
                return Dx, Dy, Dz, vrx, vry, vrz, angle, iok

    # ----------------
    #   Spherical
    # ----------------

    elif dgeom['type'] == 'spherical':

        rc = dgeom['curve_r'][0]
        rcs = np.sign(rc)
        rca = np.abs(rc)

        def ptsvect(
            # pts
            pts_x=None,
            pts_y=None,
            pts_z=None,
            # vect
            vect_x=None,
            vect_y=None,
            vect_z=None,
            # surface
            O=dgeom['cent'] + dgeom['nin'] * rc,
            rcs=rcs,
            rca=rca,
            # limits
            dthetamax=dgeom['extenthalf'][0],
            phimax=dgeom['extenthalf'][1],
            # local coordinates
            nin=dgeom['nin'],
            e0=dgeom['e0'],
            e1=dgeom['e1'],
            # return
            strict=None,
            return_x01=None,
            # timing
            dt=None,
        ):
            """

            cf. Notes_Upgrades/raytracing_surface3draytracing_surface3d.pdf

            """
            # normalize vect
            vn = np.sqrt(vect_x**2 + vect_y**2 + vect_z**2)
            vect_x = vect_x / vn
            vect_y = vect_y / vn
            vect_z = vect_z / vn

            # ------------------------------------------
            # Get local coordinates of reflection points

            # get parameters for k
            OAx = pts_x - O[0]
            OAy = pts_y - O[1]
            OAz = pts_z - O[2]
            OA2 = OAx**2 + OAy**2 + OAz**2
            OAe = OAx * vect_x + OAy * vect_y + OAz * vect_z

            iok = np.isfinite(OA2) & np.isfinite(OAe)

            # Prepare D, theta, xx
            shape = OAe.shape
            (
                Dx, Dy, Dz, vrx, vry, vrz,
                angle, phi, dtheta, kk,
            ) = _common_prepare(shape)

            # get k
            delta = (OAe**2 - (OA2-rca**2))[iok]
            ipos = delta >= 0
            iok[iok] = ipos
            sol0 = -OAe[iok] - np.sqrt(delta[ipos])
            sol1 = -OAe[iok] + np.sqrt(delta[ipos])
            kk[iok] = np.maximum(sol0, sol1)

            if np.any(iok):
                Dx = pts_x + kk*vect_x
                Dy = pts_y + kk*vect_y
                Dz = pts_z + kk*vect_z

                ODn = np.sqrt((Dx - O[0])**2 + (Dy - O[1])**2 + (Dz - O[2])**2)
                nox = rcs * (Dx[iok] - O[0]) / ODn[iok]
                noy = rcs * (Dy[iok] - O[1]) / ODn[iok]
                noz = rcs * (Dz[iok] - O[2]) / ODn[iok]

                # scalar product (for angle + reflection)
                if np.isscalar(vect_x):
                    scavn = -(vect_x*nox + vect_y*noy + vect_z*noz)

                    # get vect_reflect
                    vrx[iok] = vect_x + 2.*scavn * nox
                    vry[iok] = vect_y + 2.*scavn * noy
                    vrz[iok] = vect_z + 2.*scavn * noz

                else:
                    scavn = -(
                        vect_x[iok]*nox + vect_y[iok]*noy + vect_z[iok]*noz
                    )
                    # get vect_reflect
                    vrx[iok] = vect_x[iok] + 2.*scavn * nox
                    vry[iok] = vect_y[iok] + 2.*scavn * noy
                    vrz[iok] = vect_z[iok] + 2.*scavn * noz

                angle[iok] = -np.arcsin(scavn)

                # x0, x1
                if strict is True or return_x01 is True:
                    dtheta[iok] = rcs * np.arcsin(
                        nox*e1[0] + noy*e1[1] + noz*e1[2]
                    )
                    phi[iok] = rcs * np.arcsin(
                        (nox*e0[0] + noy*e0[1] + noz*e0[2])
                        / np.cos(dtheta[iok])
                    )

                    if strict is True:
                        iout = (
                            (np.abs(dtheta) > dthetamax)
                            | (np.abs(phi) > phimax)
                        )

                        if np.any(iout):
                            Dx[iout] = np.nan
                            Dy[iout] = np.nan
                            Dz[iout] = np.nan
                            vrx[iout] = np.nan
                            vry[iout] = np.nan
                            vrz[iout] = np.nan
                            angle[iout] = np.nan
                            iok[iout] = False

                # enforce normalization
                vnorm = np.sqrt(vrx[iok]**2 + vry[iok]**2 + vrz[iok]**2)
                vrx[iok] = vrx[iok] / vnorm
                vry[iok] = vry[iok] / vnorm
                vrz[iok] = vrz[iok] / vnorm

            # return
            if return_x01:
                return Dx, Dy, Dz, vrx, vry, vrz, angle, iok, phi, dtheta
            else:
                return Dx, Dy, Dz, vrx, vry, vrz, angle, iok

    # ----------------
    #   Toroidal
    # ----------------

    elif dgeom['type'] == 'toroidal':

        raise NotImplementedError()

    return ptsvect


# ##################################################################
# ##################################################################
#                   preparation routine
# ##################################################################


def _get_ptsvect_plane(
    plane_cent=None,
    plane_nin=None,
    plane_e0=None,
    plane_e1=None,
    # limits
    x0max=None,
    x1max=None,
):

    def ptsvect(
        pts_x=None,
        pts_y=None,
        pts_z=None,
        # pts
        vect_x=None,
        vect_y=None,
        vect_z=None,
        # surface
        cent=plane_cent,
        nin=plane_nin,
        e0=plane_e0,
        e1=plane_e1,
        # limits
        x0max=x0max,
        x1max=x1max,
        # return
        strict=None,
        return_x01=None,
    ):
        """
        """

        # ------------------------------------------
        # Get local coordinates of reflection points

        # normalize vect
        vn = np.sqrt(vect_x**2 + vect_y**2 + vect_z**2)
        vect_x = vect_x / vn
        vect_y = vect_y / vn
        vect_z = vect_z / vn

        # get parameters for k
        scavn = vect_x*nin[0] + vect_y*nin[1] + vect_z*nin[2]

        kk = (
            (
                (cent[0] - pts_x)*nin[0]
                + (cent[1] - pts_y)*nin[1]
                + (cent[2] - pts_z)*nin[2]
            )
            / scavn
        )

        # get D
        Dx = pts_x + kk * vect_x
        Dy = pts_y + kk * vect_y
        Dz = pts_z + kk * vect_z

        # get vect_reflect
        vrx = vect_x - 2.*scavn * nin[0]
        vry = vect_y - 2.*scavn * nin[1]
        vrz = vect_z - 2.*scavn * nin[2]

        # angles
        angle = -np.arcsin(scavn)

        # x0, x1
        if strict is True or return_x01 is True:
            x0 = (
                (Dx - cent[0])*e0[0]
                + (Dy - cent[1])*e0[1]
                + (Dz - cent[2])*e0[2]
            )
            x1 = (
                (Dx - cent[0])*e1[0]
                + (Dy - cent[1])*e1[1]
                + (Dz - cent[2])*e1[2]
            )

            if strict is True:
                iout = (np.abs(x0) > x0max) | (np.abs(x1) > x1max)

                if np.any(iout):
                    Dx[iout] = np.nan
                    Dy[iout] = np.nan
                    Dz[iout] = np.nan
                    vrx[iout] = np.nan
                    vry[iout] = np.nan
                    vrz[iout] = np.nan
                    angle[iout] = np.nan

        iok = np.isfinite(Dx)
        if return_x01:
            return Dx, Dy, Dz, vrx, vry, vrz, angle, iok, x0, x1
        else:
            return Dx, Dy, Dz, vrx, vry, vrz, angle, iok
    return ptsvect


# #################################################################
# #################################################################
#           Common formulas
# #################################################################


def _common_coefs(rc=None, OAz2=None, OAzez=None, ez2=None):
    return ez2, 2.*OAzez, OAz2 - rc**2


def _common_prepare(shape):
    return (
        np.full(shape, np.nan),
        np.full(shape, np.nan),
        np.full(shape, np.nan),
        np.full(shape, np.nan),
        np.full(shape, np.nan),
        np.full(shape, np.nan),
        np.full(shape, np.nan),
        np.full(shape, np.nan),
        np.full(shape, np.nan),
        np.full(shape, np.nan),
    )


def _common_kE(C0=None, C1=None, C2=None):
    kk = np.roots(np.r_[C0, C1, C2])
    kk = np.real(kk[np.isreal(kk)])

    # outgoing is necessarily the maxium k
    if np.any(kk > 0.):
        return np.max(kk[kk > 0.])
    else:
        return np.nan


# #################################################################
# #################################################################
#           Debug
# #################################################################


def _debug_cylindrical(
    pt_x=None,
    pt_y=None,
    pt_z=None,
    pts_x=None,
    pts_y=None,
    pts_z=None,
    kk=None,
    O=None,
    rc=None,
    ABx=None,
    ABy=None,
    ABz=None,
    nox=None,
    noy=None,
    noz=None,
    nin=None,
    eax=None,
    xx=None,
    theta=None,
    xmax=None,
    thetamax=None,
    check=None,
    **kwdargs,
):

    # get E
    Ex = pt_x + kk * ABx
    Ey = pt_y + kk * ABy
    Ez = pt_z + kk * ABz

    # get D
    Dx = O[0] + xx*eax[0] + rc*nox
    Dy = O[1] + xx*eax[1] + rc*noy
    Dz = O[2] + xx*eax[2] + rc*noz

    # get e1
    e1 = np.cross(nin, eax)

    import matplotlib.pyplot as plt
    fig = plt.figure()
    ax = fig.add_axes([0.1, 0.1, 0.8, 0.8], projection='3d')

    ax.plot(
        np.r_[pt_x, pts_x],
        np.r_[pt_y, pts_y],
        np.r_[pt_z, pts_z],
        '.-k',
    )

    for ii in range(len(kk)):
        l0, = ax.plot(
            np.r_[pt_x, Dx[ii], pts_x],
            np.r_[pt_y, Dy[ii], pts_y],
            np.r_[pt_z, Dz[ii], pts_z],
            '.-',
        )
        ax.plot(
            np.r_[Ex[ii], Dx[ii]],
            np.r_[Ey[ii], Dy[ii]],
            np.r_[Ez[ii], Dz[ii]],
            ls='--',
            marker='.',
            c=l0.get_color(),
        )

    # cents
    ax.plot(
        np.r_[O[0], O[0] - rc*nin[0]],
        np.r_[O[1], O[1] - rc*nin[1]],
        np.r_[O[2], O[2] - rc*nin[2]],
        'ob',
    )

    # axe
    ax.plot(
        O[0] + xmax*np.r_[-1, 1] * eax[0],
        O[1] + xmax*np.r_[-1, 1] * eax[1],
        O[2] + xmax*np.r_[-1, 1] * eax[2],
        '--k',
    )

    # circles
    ang = thetamax*np.linspace(-1, 1, 11)
    ax.plot(
        O[0] + rc*(-np.cos(ang)*nin[0] + np.sin(ang)*e1[0]),
        O[1] + rc*(-np.cos(ang)*nin[1] + np.sin(ang)*e1[1]),
        O[2] + rc*(-np.cos(ang)*nin[2] + np.sin(ang)*e1[2]),
        '--k',
    )

    # check angles
    DA = np.r_[pt_x - Dx, pt_y - Dy, pt_z - Dz]
    DB = np.r_[pts_x - Dx, pts_y - Dy, pts_z - Dz]
    DE = np.r_[Ex - Dx, Ey - Dy, Ez - Dz]
    nA = np.linalg.norm(DA)
    nB = np.linalg.norm(DB)
    nE = np.linalg.norm(DE)
    print(np.sum(DA*DE) / (nA*nE))
    print(np.sum(DB*DE) / (nB*nE))

    import pdb; pdb.set_trace()     # DB


def _debug_spherical(
    pt_x=None,
    pt_y=None,
    pt_z=None,
    pts_x=None,
    pts_y=None,
    pts_z=None,
    rr=None,
    O=None,
    rc=None,
    ABx=None,
    ABy=None,
    ABz=None,
    nin=None,
    e0=None,
    e1=None,
    **kwdargs,
):

    # get E
    Ex = pt_x + rr * ABx
    Ey = pt_y + rr * ABy
    Ez = pt_z + rr * ABz

    # get nout(E)
    nout_x = Ex - O[0]
    nout_y = Ey - O[1]
    nout_z = Ez - O[2]
    nout_norm = np.sqrt(nout_x**2 + nout_y**2 + nout_z**2)
    nout_x = nout_x / nout_norm
    nout_y = nout_y / nout_norm
    nout_z = nout_z / nout_norm

    # get D
    Dx = O[0] + rc*nout_x
    Dy = O[1] + rc*nout_y
    Dz = O[2] + rc*nout_z

    import matplotlib.pyplot as plt
    fig = plt.figure()
    ax = fig.add_axes([0.1, 0.1, 0.8, 0.8], projection='3d')

    ax.plot(
        np.r_[pt_x, pts_x],
        np.r_[pt_y, pts_y],
        np.r_[pt_z, pts_z],
        '.-k',
    )

    for ii in range(len(rr)):
        l0, = ax.plot(
            np.r_[pt_x, Dx[ii], pts_x],
            np.r_[pt_y, Dy[ii], pts_y],
            np.r_[pt_z, Dz[ii], pts_z],
            '.-',
        )
        ax.plot(
            np.r_[Ex[ii], Dx[ii]],
            np.r_[Ey[ii], Dy[ii]],
            np.r_[Ez[ii], Dz[ii]],
            ls='--',
            marker='.',
            c=l0.get_color(),
        )

    # circles
    ax.plot(
        np.r_[O[0], O[0] - rc*nin[0]],
        np.r_[O[1], O[1] - rc*nin[1]],
        np.r_[O[2], O[2] - rc*nin[2]],
        'ob',
    )

    # circles
    theta = np.pi*np.linspace(-1, 1, 100)[1:]
    ax.plot(
        O[0] + rc*(np.cos(theta)*nin[0] + np.sin(theta)*e0[0]),
        O[1] + rc*(np.cos(theta)*nin[1] + np.sin(theta)*e0[1]),
        O[2] + rc*(np.cos(theta)*nin[2] + np.sin(theta)*e0[2]),
        '--k',
    )
    ax.plot(
        O[0] + rc*(np.cos(theta)*nin[0] + np.sin(theta)*e1[0]),
        O[1] + rc*(np.cos(theta)*nin[1] + np.sin(theta)*e1[1]),
        O[2] + rc*(np.cos(theta)*nin[2] + np.sin(theta)*e1[2]),
        '--k',
    )
    ax.plot(
        O[0] + rc*(np.cos(theta)*e0[0] + np.sin(theta)*e1[0]),
        O[1] + rc*(np.cos(theta)*e0[1] + np.sin(theta)*e1[1]),
        O[2] + rc*(np.cos(theta)*e0[2] + np.sin(theta)*e1[2]),
        '--k',
    )

    # check angles
    DA = np.r_[pt_x - Dx, pt_y - Dy, pt_z - Dz]
    DB = np.r_[pts_x - Dx, pts_y - Dy, pts_z - Dz]
    DE = np.r_[Ex - Dx, Ey - Dy, Ez - Dz]
    nA = np.linalg.norm(DA)
    nB = np.linalg.norm(DB)
    nE = np.linalg.norm(DE)
    print(np.sum(DA*DE) / (nA*nE))
    print(np.sum(DB*DE) / (nB*nE))

    import pdb; pdb.set_trace()     # DB



# -*- coding: utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
# Also if needed: retab
'''
    EQUIMAP tools module, functions ...
'''
# Standard python modules
from __future__ import (unicode_literals, absolute_import,  \
                        print_function, division)
import numpy as np
import os
import re
import scipy.interpolate as interpolate
import warnings
#import sys

# Local modules
import imas
try:
    import imas_west
except ImportError as err:
    pass
    print(err)
try:
    import pywed as pw
except ImportError as err:
    pass
    print(err)

# Project modules
try:
    import mag_ripple as mr
except ImportError as err:
    pass
    print(err)

__all__ = ['get']

# Parameters
min_Ip = 100000

def get(shot, time, R, Phi, Z, quantity, no_ripple=False, \
        run=0, occ=0, user='imas_public', machine='west'):
    '''

    Interpolation of the requested quantity for the list time at
    input coodinates R, Phi, Z

    WARNING: Ripple is only taken into account for 'b_field_norm', 'b_field_r',
             'b_field_z' and b_field_tor' quantities for the moment

    Coordinate convention used (COCOS 11, see Sauter, Medvedev, Comp.Phys.Com. 184, 2013)
    -------------------------------------------------------------------------------------
     Cylindrical |     Poloidal      | Phi from top  | theta (pol ang) from front |   psi
    ------------------------------------------------------------------------------------------
     (R, Phi, Z) | (rho, theta, Phi) | cnt-clockwise |         clockwise          | increasing

    Parameters
    ----------
    shot : int
        shot number
    time : list of floats [s]
        times where to perform interpolation (ABSOLUTE TIME, without t_ignitron)
    R : list of floats [m]
        big radius where to perform interpolation
    Phi : list of floats, same length as R [rad]
        toroidal angle where to perform interpolation
    Z : list of floats, same length as R [m]
        vertical coordinate where to perform interpolation
    quantity : string
        for which quantity perform interpolation. One of

          ``rho_pol_norm``
            poloidal flux coordinate (normalized)

          ``rho_tor_norm``
            toroidal flux coordinate (normalized)

          ``rho_tor``
            toroidal flux coordinate [m]

          ``psi``
            poloidal flux [Wb]

          ``phi``
            toroidal flux [Wb]

          ``theta``
            poloidal angle [rad] in the range [0, 2*pi[

          ``j_tor``
            toroidal current density [A.m^-2]

          ``j_parallel``
            parallel current density [A.m^-2]

          ``b_field_r``
            big radius (R) component of the poloidal magnetic field [T]

          ``b_field_z``
            vertical (Z) component of the poloidal magnetic field [T]

          ``b_field_tor``
            toroidal component of the magnetic field [T]

          ``b_field_norm``
            total magnetic field norm [T]

    no_ripple : boolean, optional (default=False)
        do not calculate magnetic ripple
    run : run number, optional (default=0)
    occ : occurrence number, optional (default=0)
    user : user name, optional (default=imas_public)
    machine : machine name, optional (default=west)

    Returns
    -------
    out : ndarray, shape (time, points)
        interpolated quantity for list time at points with coordinates R, Phi, Z

    '''

    #print('time       =', time)
    #print('R          =', R)
    #print('Phi        =', Phi)
    #print('Z          =', Z)
    print('quantity   =', quantity)

    # Check if shot exists
    run_number = '{:04d}'.format(run)
    shot_file  = os.path.expanduser('~' + user + '/public/imasdb/' + machine + \
                                    '/3/0/' + 'ids_' + str(shot) + run_number + \
                                    '.datafile')
    if (not os.path.isfile(shot_file)):
        raise FileNotFoundError('IMAS file does not exist')

    # Get equilibrium fast
    idd = imas.ids(shot, run)
    idd.open_env(user, machine, '3')
    idd.equilibrium.get()
    equi = idd.equilibrium

    # Check code.output_flag for data validity
    if (np.any(np.isnan(equi.code.output_flag))):
        mask = np.full(len(equi.time), True, dtype=bool)
    else:
        mask = np.asarray(equi.code.output_flag) >= 0

    bool_b_field = False
    if isinstance(quantity, list):
        ar_bool = np.full(len(quantity), False)
        for ii in range(len(quantity)):
            if (re.match('b_field_*', quantity[ii])):
                ar_bool[ii] = True
        if (np.any(ar_bool)):
            bool_b_field = True
    else:
        if (re.match('b_field_*', quantity)):
            bool_b_field = True

    if (bool_b_field):
        # Get Itor (current of toroidal coils, coils that produce the toroidal field)
        itor, t_itor = pw.tsbase(shot, 'gmag_itor', nargout=2)
        t_ignitron   = pw.tsmat(shot, 'IGNITRON|1')
        t_itor += t_ignitron[0]

        ismag, tsmag  = pw.tsbase(shot, 'SMAG_IP', nargout=2)
        t_smag_filter = tsmag[(abs(ismag[:, 0]*1000) > min_Ip), 0]

        t_mid = 0.5*(t_smag_filter[-1] - t_smag_filter[0]) \
              + t_ignitron[0]

        ind_mid = np.abs(equi.time[mask] - t_mid).argmin()
    else:
        itor       = None
        t_itor     = None
        t_ignitron = None
        ind_mid    = None

    ar_time = np.atleast_1d(np.squeeze(np.asarray([time])))
    ar_R    = np.atleast_1d(np.squeeze(np.asarray([R])))
    ar_Phi  = np.atleast_1d(np.squeeze(np.asarray([Phi])))
    ar_Z    = np.atleast_1d(np.squeeze(np.asarray([Z])))

    if (ar_time.size > 1):
        mask_time_tmp = (equi.time[mask] >= ar_time.min()) \
                      & (equi.time[mask] <= ar_time.max())
        indMin = np.abs(equi.time[mask] \
               - equi.time[mask][mask_time_tmp][0]).argmin()
        indMax = np.abs(equi.time[mask] \
               - equi.time[mask][mask_time_tmp][-1]).argmin()
        if (indMin == 0):
            indMinApply = indMin
        else:
            indMinApply = indMin - 1
        if (indMax == (equi.time[mask].size-1)):
            indMaxApply = indMax
        else:
            indMaxApply = indMax + 1
        mask_time     = (equi.time[mask] >= equi.time[mask][indMinApply]) \
                      & (equi.time[mask] <= equi.time[mask][indMaxApply])
        time_points   = equi.time[mask][mask_time]
        #firstSpaceInterp = True # For test
        if (ar_time.size > time_points.size):
            print('__________> First perform space interpolation')
            firstSpaceInterp = True # In this case is fast to interpolate first spatially
        else:
            firstSpaceInterp = False
    else:
        firstSpaceInterp = False
        mask_time        = None
        time_points      = None

    equiDict = {}

    # Declaration of arrays 2d plots
    equi_grid = idd.equilibrium.grids_ggd[0].grid[0]
    NbrPoints = len(equi_grid.space[0].objects_per_dimension[0].object)
    equiDict['r'] = np.full(NbrPoints, np.nan)
    equiDict['z'] = np.full(NbrPoints, np.nan)
    for ii in range(NbrPoints):
        equiDict['r'][ii] = equi_grid.space[0].objects_per_dimension[0]. \
                            object[ii].geometry[0]
        equiDict['z'][ii] = equi_grid.space[0].objects_per_dimension[0]. \
                            object[ii].geometry[1]

    ind_valid = np.argmax(mask)
    NbrProf    = len(equi.time_slice[ind_valid].profiles_1d.psi)
    equiDict['psi']         = np.full((len(equi.time), NbrPoints), np.nan)
    equiDict['phi']         = np.full((len(equi.time), NbrPoints), np.nan)
    equiDict['theta']       = np.full((len(equi.time), NbrPoints), np.nan)
    equiDict['j_tor']       = np.full((len(equi.time), NbrPoints), np.nan)
    equiDict['j_parallel']  = np.full((len(equi.time), NbrPoints), np.nan)
    equiDict['b_field_r']   = np.full((len(equi.time), NbrPoints), np.nan)
    equiDict['b_field_z']   = np.full((len(equi.time), NbrPoints), np.nan)
    equiDict['b_field_tor'] = np.full((len(equi.time), NbrPoints), np.nan)
    equiDict['prof_1d_psi'] = np.full((len(equi.time), NbrProf), np.nan)
    equiDict['prof_1d_rho_tor'] = np.full((len(equi.time), NbrProf), np.nan)
    equiDict['mag_axis_r']      = np.full(len(equi.time), np.nan)
    equiDict['mag_axis_z']      = np.full(len(equi.time), np.nan)
    for ii in range(len(equi.time)):
        equi_slice  = equi.time_slice[ii]
        equi_space  = equi.time_slice[ii].ggd[0]
        if (equi_space.psi):
            equiDict['psi'][ii]         = equi_space.psi[0].values
        if (equi_space.phi):
            equiDict['phi'][ii]         = equi_space.phi[0].values
        if (equi_space.theta):
            equiDict['theta'][ii]       = equi_space.theta[0].values
        if (equi_space.j_tor):
            equiDict['j_tor'][ii]       = equi_space.j_tor[0].values
        if (equi_space.j_parallel):
            equiDict['j_parallel'][ii]  = equi_space.j_parallel[0].values
        if (equi_space.b_field_r):
            equiDict['b_field_r'][ii]   = equi_space.b_field_r[0].values
        if (equi_space.b_field_z):
            equiDict['b_field_z'][ii]   = equi_space.b_field_z[0].values
        if (equi_space.b_field_tor):
            equiDict['b_field_tor'][ii] = equi_space.b_field_tor[0].values
        equiDict['prof_1d_psi'][ii]     = equi_slice.profiles_1d.psi
        equiDict['prof_1d_rho_tor'][ii] = equi_slice.profiles_1d.rho_tor
        equiDict['mag_axis_r'][ii]      = equi_slice.global_quantities.magnetic_axis.r
        equiDict['mag_axis_z'][ii]      = equi_slice.global_quantities.magnetic_axis.z

    points        = np.vstack((equiDict['r'], equiDict['z'])).transpose()
    interp_points = np.vstack((ar_R, ar_Z)).transpose()

    if isinstance(quantity, list):
        out = {}
        for iquant in quantity:
            out[iquant] = \
              interp_quantity(iquant, interp_points, points, time_points, equi, \
                              ar_time, ar_R, ar_Phi, ar_Z, mask, mask_time, \
                              firstSpaceInterp, itor, t_itor, t_ignitron, \
                              no_ripple, ind_mid, equiDict)
    else:
        out = interp_quantity(quantity, interp_points, points, time_points, equi, \
                              ar_time, ar_R, ar_Phi, ar_Z, mask, mask_time, \
                              firstSpaceInterp, itor, t_itor, t_ignitron, \
                              no_ripple, ind_mid, equiDict)
    return out


def interp_quantity(quantity, interp_points, points, time_points, equi, \
                    ar_time, ar_R, ar_Phi, ar_Z, mask, mask_time, \
                    firstSpaceInterp, itor, t_itor, t_ignitron, \
                    no_ripple, ind_mid, equiDict):

    value_interpolated = np.full((ar_time.size, ar_R.size), np.nan)
    if (firstSpaceInterp):
        value_interpSpace  = np.full((time_points.size, ar_R.size), np.nan)

    # Computation of requested quantities
    if (quantity == 'b_field_norm'):
        if (no_ripple):
            b_field_norm = np.sqrt(equiDict['b_field_r']**2. \
                                 + equiDict['b_field_z']**2. \
                                 + equiDict['b_field_tor']**2.)
            if (firstSpaceInterp):
                # Space interpolation
                for ii in range(time_points.size):
                    lin_intp = interpolate.LinearNDInterpolator(points, \
                                 b_field_norm[mask, :][mask_time][ii, :])
                    value_interpSpace[ii, :] = lin_intp.__call__(interp_points)
            else:
                # Time interpolation
                f_intp = interpolate.interp1d(equi.time[mask], \
                                              b_field_norm[mask, :], axis=0, \
                                              bounds_error=False)
        else: # B_norm with ripple calculation
            # Declaration arrays
            br_intp = np.full((ar_time.size, ar_R.size), np.nan)
            bt_intp = np.full((ar_time.size, ar_R.size), np.nan)
            bz_intp = np.full((ar_time.size, ar_R.size), np.nan)

            if (firstSpaceInterp):
                br_Sintp = np.full((time_points.size, ar_R.size), np.nan)
                bt_Sintp = np.full((time_points.size, ar_R.size), np.nan)
                bz_Sintp = np.full((time_points.size, ar_R.size), np.nan)
                # Space interpolation
                for ii in range(time_points.size):
                    lin_intp = interpolate.LinearNDInterpolator(points, \
                               equiDict['b_field_r'][mask, :][mask_time][ii, :])
                    br_Sintp[ii, :] = lin_intp.__call__(interp_points)
                    lin_intp = interpolate.LinearNDInterpolator(points, \
                               equiDict['b_field_tor'][mask, :][mask_time][ii, :])
                    bt_Sintp[ii, :] = lin_intp.__call__(interp_points)
                    lin_intp = interpolate.LinearNDInterpolator(points, \
                               equiDict['b_field_z'][mask, :][mask_time][ii, :])
                    bz_Sintp[ii, :] = lin_intp.__call__(interp_points)
                # Time interpolation
                f_intp = interpolate.interp1d(time_points, \
                                              br_Sintp, axis=0, \
                                              bounds_error=False)
                br_intp = np.atleast_2d(np.squeeze(f_intp(ar_time)))

                f_intp = interpolate.interp1d(time_points, \
                                              bt_Sintp, axis=0, \
                                              bounds_error=False)
                bt_intp = np.atleast_2d(np.squeeze(f_intp(ar_time)))

                f_intp = interpolate.interp1d(time_points, \
                                              bz_Sintp, axis=0, \
                                              bounds_error=False)
                bz_intp = np.atleast_2d(np.squeeze(f_intp(ar_time)))
            else:
                # Time interpolation
                f_intp_br = interpolate.interp1d(equi.time[mask], \
                              equiDict['b_field_r'][mask, :], axis=0, \
                              bounds_error=False)
                f_intp_bt = interpolate.interp1d(equi.time[mask], \
                              equiDict['b_field_tor'][mask, :], axis=0, \
                              bounds_error=False)
                f_intp_bz = interpolate.interp1d(equi.time[mask], \
                              equiDict['b_field_z'][mask, :], axis=0, \
                              bounds_error=False)

                br_intp_t = np.atleast_2d(np.squeeze(f_intp_br(ar_time)))
                bt_intp_t = np.atleast_2d(np.squeeze(f_intp_bt(ar_time)))
                bz_intp_t = np.atleast_2d(np.squeeze(f_intp_bz(ar_time)))
                # Space interpolation
                for ii in range(ar_time.size):
                    lin_intp = interpolate.LinearNDInterpolator(points, br_intp_t[ii])
                    br_intp[ii, :] = lin_intp.__call__(interp_points)

                    lin_intp = interpolate.LinearNDInterpolator(points, bt_intp_t[ii])
                    bt_intp[ii, :] = lin_intp.__call__(interp_points)

                    lin_intp = interpolate.LinearNDInterpolator(points, bz_intp_t[ii])
                    bz_intp[ii, :] = lin_intp.__call__(interp_points)

            # Interpolate current
            itor_intp_t = np.interp(ar_time, t_itor[:, 0], itor[:, 0])
            b0_intp_t   = np.interp(ar_time, equi.time[mask], \
                                    equi.vacuum_toroidal_field.b0[mask])

            # Compute reference vaccuum magnetic field
            bt_vac = equi.vacuum_toroidal_field.r0*b0_intp_t[:, np.newaxis] \
                   / ar_R[np.newaxis, :]

            # Compute magnetic field for given Phi
            br_ripple, bt_ripple, bz_ripple = mr.mag_ripple(ar_R, ar_Phi, \
                                                 ar_Z, itor_intp_t)

            # Check and correct Br if needed
            r_ax = equi.time_slice[ind_mid].global_quantities.magnetic_axis.r

            z_mid_ax = \
              equi.time_slice[ind_mid].global_quantities.magnetic_axis.z \
                     + 0.5*equi.time_slice[ind_mid].boundary.minor_radius

            ind_p = np.abs((equiDict['r'] - r_ax)**2. \
                         + (equiDict['z'] - z_mid_ax)**2).argmin()

            if (equiDict['b_field_r'][ind_mid, ind_p] > 0):
                br_intp *= -1
                warnings.warn('Correcting b_field_r in b_field_norm, for negative toroidal current COCOS 11')

            # Check and correct Bz if needed
            r_mid_ax = \
              equi.time_slice[ind_mid].global_quantities.magnetic_axis.r \
                     + 0.5*equi.time_slice[ind_mid].boundary.minor_radius
            z_ax = equi.time_slice[ind_mid].global_quantities.magnetic_axis.z

            ind_p = np.abs((equiDict['r'] - r_mid_ax)**2. \
                         + (equiDict['z'] - z_ax)**2).argmin()

            if (equiDict['b_field_z'][ind_mid, ind_p] < 0):
                bz_intp *= -1
                warnings.warn('Correcting b_field_z in b_field_norm, for negative toroidal current COCOS 11')

            # Check and correct Btor if needed
            ind_p = np.abs((equiDict['r'] - r_ax)**2. \
                         + (equiDict['z'] - z_ax)**2).argmin()

            if (equiDict['b_field_tor'][ind_mid, ind_p] > 0):
                bt_intp *= -1
                warnings.warn('Correcting b_field_tor in b_field_norm, for negative toroidal field COCOS 11')

            # Value interpolated
            value_interpolated = \
              np.sqrt((br_intp - br_ripple)**2. \
                    + (np.abs(bt_intp - bt_ripple) - np.abs(bt_vac))**2. \
                    + (bz_intp - bz_ripple)**2.)

            return np.squeeze(value_interpolated)

    elif (not no_ripple and re.match('b_field_*', quantity)):
            # Declaration arrays
            br_ripple = np.full((ar_time.size, ar_R.size), np.nan)
            bt_ripple = np.full((ar_time.size, ar_R.size), np.nan)
            bz_ripple = np.full((ar_time.size, ar_R.size), np.nan)

            if (firstSpaceInterp):
                # Space interpolation
                quant_mask = eval('equiDict["'+quantity+'"][mask, :][mask_time]')
                for ii in range(time_points.size):
                    lin_intp = interpolate.LinearNDInterpolator(points, \
                                                      quant_mask[ii, :])
                    value_interpSpace[ii, :] = lin_intp.__call__(interp_points)

                # Time interpolation
                f_intp = interpolate.interp1d(time_points, \
                                              value_interpSpace, axis=0, \
                                              bounds_error=False)
                value_interpolated = np.atleast_2d(np.squeeze(f_intp(ar_time)))
            else:
                # Time interpolation
                f_intp = interpolate.interp1d(equi.time[mask], \
                              eval('equiDict["'+quantity+'"][mask, :]'), axis=0, \
                              bounds_error=False)
                b_intp_t = np.atleast_2d(np.squeeze(f_intp(ar_time)))

                # Space interpolation
                for ii in range(ar_time.size):
                    lin_intp = interpolate.LinearNDInterpolator(points, b_intp_t[ii])
                    value_interpolated[ii, :] = lin_intp.__call__(interp_points)

            # Interpolate current
            itor_intp_t = np.interp(ar_time, t_itor[:, 0], itor[:, 0])

            # Compute magnetic field for given Phi
            br_ripple, bt_ripple, bz_ripple = mr.mag_ripple(ar_R, ar_Phi, \
                                                 ar_Z, itor_intp_t)

            if (quantity == 'b_field_r'):
                r_ax = equi.time_slice[ind_mid].global_quantities.magnetic_axis.r

                z_mid_ax = \
                  equi.time_slice[ind_mid].global_quantities.magnetic_axis.z \
                         + 0.5*equi.time_slice[ind_mid].boundary.minor_radius

                ind_p = np.abs((equiDict['r'] - r_ax)**2. \
                             + (equiDict['z'] - z_mid_ax)**2).argmin()

                if (equiDict['b_field_r'][ind_mid, ind_p] > 0):
                    value_interpolated *= -1
                    value_interpolated -= br_ripple
                    warnings.warn('Correcting b_field_r, for negative toroidal current COCOS 11')
                else:
                    value_interpolated -= br_ripple

            elif (quantity == 'b_field_z'):
                r_mid_ax = \
                  equi.time_slice[ind_mid].global_quantities.magnetic_axis.r \
                         + 0.5*equi.time_slice[ind_mid].boundary.minor_radius
                z_ax = equi.time_slice[ind_mid].global_quantities.magnetic_axis.z

                ind_p = np.abs((equiDict['r'] - r_mid_ax)**2. \
                             + (equiDict['z'] - z_ax)**2).argmin()

                if (equiDict['b_field_z'][ind_mid, ind_p] < 0):
                    value_interpolated *= -1
                    value_interpolated -= bz_ripple
                    warnings.warn('Correcting b_field_z, for negative toroidal current COCOS 11')
                else:
                    value_interpolated -= bz_ripple

            elif (quantity == 'b_field_tor'):
                b0_intp_t = np.interp(ar_time, equi.time[mask], \
                                      equi.vacuum_toroidal_field.b0[mask])
                # Compute reference vaccuum magnetic field
                bt_vac = equi.vacuum_toroidal_field.r0*b0_intp_t[:, np.newaxis] \
                       / ar_R[np.newaxis, :]

                r_ax = equi.time_slice[ind_mid].global_quantities.magnetic_axis.r
                z_ax = equi.time_slice[ind_mid].global_quantities.magnetic_axis.z

                ind_p = np.abs((equiDict['r'] - r_ax)**2. \
                             + (equiDict['z'] - z_ax)**2).argmin()

                if (equiDict['b_field_tor'][ind_mid, ind_p] > 0):
                    value_interpolated *= -1
                    value_interpolated -= (bt_ripple - bt_vac)
                    warnings.warn('Correcting b_field_tor, for negative toroidal field COCOS 11')
                else:
                    value_interpolated -= (bt_ripple - bt_vac)
            else:
                print()
                print('ERROR: not valid quantity input:', quantity)
                print()
                raise SyntaxError

            return np.squeeze(value_interpolated)

    elif (quantity == 'rho_pol_norm'):
        rho_pol_norm = np.sqrt((equiDict['psi'] \
                     - equiDict['prof_1d_psi'][:, 0, np.newaxis]) \
                / (equiDict['prof_1d_psi'][:, -1, np.newaxis] \
                 - equiDict['prof_1d_psi'][:, 0, np.newaxis]))
        if (firstSpaceInterp):
            # Space interpolation
            for ii in range(time_points.size):
                lin_intp = interpolate.LinearNDInterpolator(points, \
                             rho_pol_norm[mask, :][mask_time][ii, :])
                value_interpSpace[ii, :] = lin_intp.__call__(interp_points)
        else:
            # Time interpolation
            f_intp = interpolate.interp1d(equi.time[mask], rho_pol_norm[mask, :], \
                                          axis=0, bounds_error=False)

    elif (quantity == 'rho_tor_norm' or quantity == 'rho_tor'):
        if (firstSpaceInterp):
            # Space interpolation
            for ii in range(time_points.size):
                lin_intp = interpolate.LinearNDInterpolator(points, \
                             equiDict['psi'][mask, :][mask_time][ii, :])
                value_interpSpace[ii, :] = lin_intp.__call__(interp_points)
        else:
            # Time interpolation
            f_intp = interpolate.interp1d(equi.time[mask], equiDict['psi'][mask, :], \
                                          axis=0, bounds_error=False)

    elif (quantity == 'theta'):
        mag_ax_r = np.atleast_1d(np.squeeze(np.interp(ar_time, \
                     equi.time[mask], equiDict['mag_axis_r'][mask])))
        mag_ax_z = np.atleast_1d(np.squeeze(np.interp(ar_time, \
                     equi.time[mask], equiDict['mag_axis_z'][mask])))

        for ii in range(ar_time.size):
            delta_R = ar_R - mag_ax_r[ii, np.newaxis]
            delta_Z = ar_Z - mag_ax_z[ii, np.newaxis]

            val_arctan = np.arctan(delta_Z/delta_R)

            mask_theta = (delta_R >= 0) & (delta_Z >= 0)
            value_interpolated[ii, mask_theta] = val_arctan[mask_theta]

            mask_theta = (delta_R >= 0) & (delta_Z < 0)
            value_interpolated[ii, mask_theta] = 2*np.pi + val_arctan[mask_theta]

            mask_theta = (delta_R < 0)
            value_interpolated[ii, mask_theta] = np.pi + val_arctan[mask_theta]

        return np.squeeze(2.*np.pi - value_interpolated)
    else:
        if (firstSpaceInterp):
            # Space interpolation
            for ii in range(time_points.size):
                lin_intp = interpolate.LinearNDInterpolator(points, \
                  eval('equiDict["'+quantity+'"][mask, :][mask_time][ii, :]'))
                value_interpSpace[ii, :] = lin_intp.__call__(interp_points)
        else:
            # Time interpolation
            f_intp = interpolate.interp1d(equi.time[mask], \
                       eval('equiDict["'+quantity+'"][mask, :]'), \
                       axis=0, bounds_error=False)

    if (firstSpaceInterp):
        # Time interpolation
        f_intp = interpolate.interp1d(time_points, \
                                      value_interpSpace, axis=0, \
                                      bounds_error=False)
        value_interpolated = np.atleast_2d(np.squeeze(f_intp(ar_time)))
    else:
        # Time interpolation
        out_time_interp = np.atleast_2d(np.squeeze(f_intp(ar_time)))
        # Space interpolation
        for ii in range(ar_time.size):
            lin_intp = interpolate.LinearNDInterpolator(points, out_time_interp[ii])
            value_interpolated[ii, :] = lin_intp.__call__(interp_points)

    # Extra calculations for rho_tor, rho_tor_norm, b_field_r and b_field_z
    if (quantity == 'rho_tor_norm' or quantity == 'rho_tor'):

        value_interp_rho_tor = np.full(value_interpolated.shape, np.nan)

        if (quantity == 'rho_tor_norm'):
            try:
                rho_tor_norm = equiDict['prof_1d_rho_tor'] \
                             / equiDict['prof_1d_rho_tor'][:, -1, np.newaxis]
            except ZeroDivisionError as err:
                print('Division by zero for rho_tor_norm interpolation:', err)
                raise
            f_intp_rho_tor = interpolate.interp1d(equi.time[mask], \
                               rho_tor_norm[mask, :], axis=0, bounds_error=False)
        elif (quantity == 'rho_tor'):
            f_intp_rho_tor = interpolate.interp1d(equi.time[mask], \
                                         equiDict['prof_1d_rho_tor'][mask, :], \
                                         axis=0, bounds_error=False)

        rho_tor1D = np.atleast_2d(np.squeeze(f_intp_rho_tor(ar_time)))

        f_intp_psi = interpolate.interp1d(equi.time[mask], \
                       equiDict['prof_1d_psi'][mask, :], axis=0, bounds_error=False)
        psi1D = np.atleast_2d(np.squeeze(f_intp_psi(ar_time)))

        for ii in range(ar_time.size):
            f_intp_psi_rho_tor = interpolate.interp1d(psi1D[ii], rho_tor1D[ii], \
                                                      bounds_error=False)
            value_interp_rho_tor[ii] = f_intp_psi_rho_tor(value_interpolated[ii])

        return np.squeeze(value_interp_rho_tor)

    elif (quantity == 'b_field_r'):
        r_ax = equi.time_slice[ind_mid].global_quantities.magnetic_axis.r

        z_mid_ax = equi.time_slice[ind_mid].global_quantities.magnetic_axis.z \
                 + 0.5*equi.time_slice[ind_mid].boundary.minor_radius

        ind_p = np.abs((equiDict['r'] - r_ax)**2. \
                     + (equiDict['z'] - z_mid_ax)**2).argmin()

        if (equiDict['b_field_r'][ind_mid, ind_p] > 0):
            value_interpolated *= -1
            warnings.warn('Correcting b_field_r, if Ip negative COCOS 11')

        return np.squeeze(value_interpolated)

    elif (quantity == 'b_field_z'):
        r_mid_ax = equi.time_slice[ind_mid].global_quantities.magnetic_axis.r \
                 + 0.5*equi.time_slice[ind_mid].boundary.minor_radius
        z_ax = equi.time_slice[ind_mid].global_quantities.magnetic_axis.z

        ind_p = np.abs((equiDict['r'] - r_mid_ax)**2. \
                     + (equiDict['z'] - z_ax)**2).argmin()

        if (equiDict['b_field_z'][ind_mid, ind_p] < 0):
            value_interpolated *= -1
            warnings.warn('Correcting b_field_z, if Ip negative COCOS 11')

        return np.squeeze(value_interpolated)

    elif (quantity == 'b_field_tor'):
        r_ax = equi.time_slice[ind_mid].global_quantities.magnetic_axis.r
        z_ax = equi.time_slice[ind_mid].global_quantities.magnetic_axis.z

        ind_p = np.abs((equiDict['r'] - r_ax)**2. \
                     + (equiDict['z'] - z_ax)**2).argmin()

        if (equiDict['b_field_tor'][ind_mid, ind_p] > 0):
            value_interpolated *= -1
            warnings.warn('Correcting b_field_tor, if b_field_tor negative COCOS 11')

        return np.squeeze(value_interpolated)
    else:
        return np.squeeze(value_interpolated)



# -*- coding: utf-8 -*-


# Built-in
# import copy


# Common
import numpy as np
import datastock as ds


# tofu
from ._class05_Crystal import Crystal as Previous
from . import _class3_check
# from . import _class6_check as _check
# from . import _class6_compute as _compute


__all__ = ['Grating']


# #############################################################################
# #############################################################################
#                           Diagnostic
# #############################################################################


class Grating(Previous):

    _dshow = dict(Previous._dshow)
    _dshow.update({
        'crystal': [
            'dgeom.type',
            'dgeom.curve_r',
            'dgeom.area',
            'dmat.name',
            'dmat.miller',
            'dgeom.outline',
            'dgeom.poly',
            'dgeom.cent',
            # 'dmisc.color',
        ],
    })

    def add_grating(
        self,
        key=None,
        # geometry
        dgeom=None,
        # material
        dmat=None,
        # dmisc
        color=None,
    ):
        """ Add a crystal

        Can be defined from:
            - 2d outline + 3d center + unit vectors (nin, e0, e1)
            - 3d polygon + nin

        Unit vectors will be checked and normalized
        If planar, area will be computed
        Outline will be made counter-clockwise

        """

        # check / format input
        dref, ddata, dobj = _class3_check._add_surface3d(
            coll=self,
            key=key,
            which='grating',
            which_short='grat',
            # 2d outline
            **dgeom,
        )

        key = list(dobj['grating'].keys())[0]

        # material
        dobj['grating'][key]['dmat'] = _check._dmat(
            dgeom=dobj['grating'][key]['dgeom'],
            dmat=dmat,
            alpha=alpha,
            beta=beta,
        )

        # dmisc
        dobj['grating'][key]['dmisc'] = _class3_check._dmisc(
            key=key,
            color=color,
        )

        # update dicts
        self.update(dref=dref, ddata=ddata, dobj=dobj)



#!/usr/bin/env/python
# coding=utf-8

import os
import subprocess

_HERE = os.path.abspath(os.path.dirname(__file__))


def updateversion(path=_HERE):
    # Fetch version from git tags, and write to version.py
    # Also, when git is not available (PyPi package), use stored version.py
    version_py = os.path.join(path, 'tofu', 'version.py')
    try:
        version_git = subprocess.check_output(["git",
                                               "describe"]).rstrip().decode()
    except subprocess.CalledProcessError:
        with open(version_py, 'r') as fh:
            version_git = fh.read().strip().split("=")[-1].replace("'", '')
    version_git = version_git.lower().replace('v', '').replace(' ', '')

    version_msg = "# Do not edit, pipeline versioning governed by git tags!"
    with open(version_py, "w") as fh:
        msg = "{0}__version__ = '{1}'{0}".format(os.linesep, version_git)
        fh.write(version_msg + msg)
    return version_git



# -*- coding: utf-8 -*-


# Built-in
import time


# Common
import numpy as np
import scipy.sparse as scpsp


import datastock as ds


# tofu
from . import _generic_check
from . import _class8_compute_signal
from . import _class10_checks as _checks
from . import _class10_algos as _algos
tomotok2tofu = _checks.tomotok2tofu


__all__ = ['get_available_inversions_algo']


# ##################################################################
# ##################################################################
#                           main
# ##################################################################


def compute_inversions(
    # resources
    coll=None,
    # inversion name
    key=None,
    # input data
    key_matrix=None,
    key_data=None,
    key_sigma=None,
    sigma=None,
    # constraints
    dconstraints=None,
    # regularity operator
    operator=None,
    geometry=None,
    # choice of algo
    algo=None,
    # misc
    solver=None,
    conv_crit=None,
    chain=None,
    verb=None,
    store=None,
    # algo and solver-specific options
    kwdargs=None,
    method=None,
    options=None,
):

    # -------------
    # check inputs

    (
        key_matrix,
        key_diag, key_data, key_sigma,
        keybs, keym, mtype,
        ddata, dsigma, matrix, units_gmat,
        keyt, t, reft, notime,
        m3d, indok, iokt,
        dconstraints,
        opmat, operator, geometry,
        dalgo, dconstraints, dcon,
        conv_crit, crop, chain, kwdargs, method, options,
        solver, verb, store,
        keyinv, refinv, regul,
    ) = _checks._compute_check(**locals())

    data = ddata['data']
    sigma = dsigma['data']

    nt, nchan = data.shape
    nbs = matrix.shape[-1]

    # -------------
    # get func

    if isinstance(dalgo['func'], str):
        if dalgo['source'] == 'tofu':
            dalgo['func'] = getattr(_algos, dalgo['func'])
        elif dalgo['source'] == 'tomotok':
            dalgo['func'] = getattr(tomotok2tofu, dalgo['func'])

    # -------------
    # prepare data

    if verb >= 1:
        # t0 = time.process_time()
        t0 = time.perf_counter()
        print("Preparing data... ", end='', flush=True)

    # indt (later)

    # normalization
    data_n = (data / sigma)
    mu0 = 1.

    # Define Regularization operator
    if dalgo['family'] == 'Non-regularized':
        R = None
    else:
        if dalgo['source'] == 'tomotok' and dalgo['reg_operator'] == 'MinFisher':
            R = opmat
        elif operator == 'D0N2':
            R = opmat[0]
        elif operator == 'D1N2':
            R = opmat[0] + opmat[1]
        elif operator == 'D2N2':
            R = opmat[0] + opmat[1]
        else:
            msg = 'unknown operator!'
            raise Exception(msg)

    # prepare computation intermediates
    precond = None
    Tyn = np.full((nbs,), np.nan)
    mat0 = matrix[0, ...] if m3d else matrix
    if dalgo['sparse'] is True:
        Tn = scpsp.diags(1./np.nanmean(sigma, axis=0)).dot(mat0)
        TTn = Tn.T.dot(Tn)
        if solver != 'spsolve':
            # preconditioner to approx inv(TTn + reg*Rn)
            # should improves convergence, but actually slower...
            # precond = scpsp.linalg.inv(TTn + mu0*R)
            pass

    else:
        Tn = mat0 / np.nanmean(sigma, axis=0)[:, None]
        TTn = Tn.T.dot(Tn)

    # prepare output arrays
    sol = np.full((nt, nbs), np.nan)
    mu = np.full((nt,), np.nan)
    chi2n = np.full((nt,), np.nan)
    regularity = np.full((nt,), np.nan)
    niter = np.zeros((nt,), dtype=int)
    spec = [None for ii in range(nt)]

    # -------------
    # initial guess

    if indok is None:
        sol0 = np.full((nbs,), np.mean(data[0, :]) / mat0.mean())
    else:
        sol0 = np.full((nbs,), np.mean(data[0, indok[0, :]]) / mat0.mean())

    if verb >= 1:
        # t1 = time.process_time()
        t1 = time.perf_counter()
        print(f"{t1-t0} s", end='\n', flush=True)
        print("Setting inital guess... ", end='', flush=True)

    # -------------
    # compute

    if verb >= 1:
        # t2 = time.process_time()
        t2 = time.perf_counter()
        print(f"{t2-t1} s", end='\n', flush=True)
        print("Starting time loop...", end='\n', flush=True)

    if dalgo['source'] == 'tofu':
        out = _compute_inv_loop(
            sol0=sol0,
            mu0=mu0,
            matrix=matrix,
            Tn=Tn,              # normalized geometry matrix (T)
            TTn=TTn,            # normalized tTT
            Tyn=Tyn,            # normalized
            R=R,                # Regularity operator
            precond=precond,
            data_n=data_n,      # normalized data
            sigma=sigma,
            indok=indok,
            # parameters
            m3d=m3d,
            dalgo=dalgo,
            isotropic=dalgo['isotropic'],
            conv_crit=conv_crit,
            sparse=dalgo['sparse'],
            positive=dalgo['positive'],
            chain=chain,
            verb=verb,
            kwdargs=kwdargs,
            method=method,
            options=options,
            dcon=dcon,
            regul=regul,
            # output
            sol=sol,
            mu=mu,
            chi2n=chi2n,
            regularity=regularity,
            niter=niter,
            spec=spec,
        )

    elif dalgo['source'] == 'tomotok':
        out = _compute_inv_loop_tomotok(
            sol0=sol0,
            mu0=mu0,
            matrix=matrix,
            Tn=Tn,              # normalized geometry matrix (T)
            TTn=TTn,            # normalized tTT
            Tyn=Tyn,            # normalized
            R=R,                # Regularity operator
            precond=precond,
            data_n=data_n,      # normalized data
            sigma=sigma,
            indok=indok,
            # parameters
            m3d=m3d,
            dalgo=dalgo,
            isotropic=dalgo['isotropic'],
            conv_crit=conv_crit,
            sparse=dalgo['sparse'],
            positive=dalgo['positive'],
            chain=chain,
            verb=verb,
            kwdargs=kwdargs,
            method=method,
            options=options,
            dcon=dcon,
            regul=regul,
            # output
            sol=sol,
            mu=mu,
            chi2n=chi2n,
            regularity=regularity,
            niter=niter,
            spec=spec,
        )

    if verb >= 1:
        # t3 = time.process_time()
        t3 = time.perf_counter()
        print(f"{t3-t2} s", end='\n', flush=True)
        print("Post-formatting results...", end='\n', flush=True)

    # ---------------------------------------------
    # estimate relative regularity for polar mesh of not regul

    if not regul and mtype == 'polar':
        clas = coll.dobj['bsplines'][keybs]['class']

        if clas.knotsa is None:
            # estimate 1d squared gradient
            kr = coll.dobj[coll._which_mesh][keym]['knots'][0]
            rr = coll.ddata[kr]['data']
            regularity = np.nansum(
                clas(
                    radius=np.linspace(rr[0], rr[-1], rr.size*10),
                    coefs=sol,
                    radius_vs_time=False,
                    deriv=1,
                )**2,
                axis=1,
            )

    # -------------
    # format output

    # reshape solution
    if crop is True:
        shapebs = coll.dobj['bsplines'][keybs]['shape']
        shapesol = tuple(np.r_[nt, shapebs])
        sol_full = np.zeros(shapesol, dtype=float)
        cropbs = coll.ddata[coll.dobj['bsplines'][keybs]['crop']]['data']
        cropbsflat = cropbs.ravel(order='F')
        iR = np.tile(np.arange(0, shapebs[0]), shapebs[1])[cropbsflat]
        iZ = np.repeat(np.arange(0, shapebs[1]), shapebs[0])[cropbsflat]
        sol_full[:, iR, iZ] = sol
    else:
        sol_full = sol

    # -------------
    # store

    if store is True:
        units = ddata['units'] / units_gmat
        key_data = ddata['keys']
        _store(**locals())

    else:
        return sol_full, mu, chi2n, regularity, niter, spec, t


def _store(
    coll=None,
    sol_full=None,
    notime=None,
    iokt=None,
    chi2n=None,
    mu=None,
    regularity=None,
    niter=None,
    t=None,
    keyinv=None,
    refinv=None,
    reft=None,
    keyt=None,
    key_diag=None,
    key_data=None,
    key_sigma=None,
    key_matrix=None,
    operator=None,
    geometry=None,
    dalgo=None,
    solver=None,
    chain=None,
    conv_crit=None,
    units=None,
    **kwdargs,
):

    # ---------------------------
    # reshape if unique time step

    if notime:
        assert sol_full.shape[0] == 1
        sol_full = sol_full[0, ...]
    else:
        # restore full size
        sol_full, chi2n, mu, regularity, niter = _restore_fullt(
            iokt=iokt,
            sol_full=sol_full,
            chi2n=chi2n,
            mu=mu,
            regularity=regularity,
            niter=niter,
        )
        nt = t.size

    # --------------------
    # Build dict

    # ddata
    ddata = {
        keyinv: {
            'data': sol_full,
            'ref': refinv,
            'units': units,
        },
    }

    dref = None
    if notime is False:
        if keyt is None:
            dref = {
                reft: {'size': nt},
            }
            ddata.update({
                f'{keyinv}-t': {
                    'data': t,
                    'ref': reft,
                    'dim': 'time',
                },
            })

        ddata.update({
            f'{keyinv}-chi2n': {
                'data': chi2n,
                'ref': reft,
            },
            f'{keyinv}-mu': {
                'data': mu,
                'ref': reft,
            },
            f'{keyinv}-reg': {
                'data': regularity,
                'ref': reft,
            },
            f'{keyinv}-niter': {
                'data': niter,
                'ref': reft,
            },
        })

    # add synthetic data
    kretro = f'{keyinv}-retro'

    # add inversion
    dobj = {
        'inversions': {
            keyinv: {
                'retrofit': kretro,
                'data_in': key_data,
                'sigma_in': key_sigma,
                'matrix': key_matrix,
                'sol': keyinv,
                'operator': operator,
                'geometry': geometry,
                'isotropic': dalgo['isotropic'],
                'algo': dalgo['name'],
                'solver': solver,
                'chain': chain,
                'positive': dalgo['positive'],
                'conv_crit': conv_crit,
            },
        },
    }

    # adjust for time
    if notime is True:
        dobj['inversions'][keyinv].update({
            'chi2n': chi2n,
            'mu': mu,
            'reg': regularity,
            'niter': niter,
        })

    # update instance
    coll.update(dobj=dobj, dref=dref, ddata=ddata)

    # add synthetic data
    keyt = coll.get_time(key=keyinv)[3]
    data_synth = coll.add_retrofit_data(
        key=kretro,
        key_diag=key_diag,
        key_matrix=key_matrix,
        key_profile2d=keyinv,
        t=keyt,
        store=True,
    )


def _restore_fullt(
    iokt=None,
    sol_full=None,
    chi2n=None,
    mu=None,
    regularity=None,
    niter=None,
):

    # sol_full
    shape = tuple(np.r_[iokt.size, sol_full.shape[1:]])
    sol_fulli = np.full(shape, np.nan)
    sol_fulli[iokt, :] = sol_full

    # 1d
    chi2ni = np.full((iokt.size,), np.nan)
    mui = np.full((iokt.size,), np.nan)
    regularityi = np.full((iokt.size,), np.nan)
    niteri = np.full((iokt.size,), np.nan)

    chi2ni[iokt] = chi2n
    mui[iokt] = mu
    regularityi[iokt] = regularity
    niteri[iokt] = niter

    return sol_fulli, chi2ni, mui, regularityi, niteri


# ##################################################################
# ##################################################################
#                   _compute time loop
# ##################################################################


def _compute_inv_loop(
    # inputs
    dalgo=None,
    sol0=None,
    mu0=None,
    matrix=None,
    Tn=None,        # normalized geometry matrix (T)
    TTn=None,       # normalized tTT
    Tyn=None,       # normalized tTy
    R=None,
    precond=None,
    data_n=None,
    sigma=None,
    indok=None,
    # parameters
    m3d=None,
    conv_crit=None,
    isotropic=None,
    sparse=None,
    positive=None,
    chain=None,
    verb=None,
    kwdargs=None,
    method=None,
    options=None,
    dcon=None,
    regul=None,
    # output
    sol=None,
    mu=None,
    chi2n=None,
    regularity=None,
    niter=None,
    spec=None,
):

    # -----------------------------------
    # Getting initial solution - step 1/2

    nt, nchan = data_n.shape
    nbs = Tn.shape[1]

    if verb >= 2:
        form = "nchan * chi2n   +   mu *  R           "
        verb2head = f"\n\t\titer    {form} \t\t\t  tau  \t      conv"
    else:
        verb2head = None

    # -----------------------------------
    # Options for quadratic solvers only

    bounds = None
    func_val, func_jac, func_hess = None, None, None
    if regul and positive is True:

        bounds = tuple([(0., None) for ii in range(0, sol0.size)])

        def func_val(x, mu=mu0, Tn=Tn, yn=data_n[0, :], TTn=None, Tyn=None):
            return np.sum((Tn.dot(x) - yn)**2) + mu*x.dot(R.dot(x))

        def func_jac(x, mu=mu0, Tn=None, yn=None, TTn=TTn, Tyn=Tyn):
            return 2.*(TTn + mu*R).dot(x) - 2.*Tyn

        def func_hess(x, mu=mu0, Tn=None, yn=None, TTn=TTn, Tyn=Tyn):
            return 2.*(TTn + mu*R)

    elif not regul:

        if positive is True:
            bounds = (
                np.zeros((nbs,), dtype=float),
                np.full((nbs,), np.inf),
            )
        else:
            bounds = (-np.inf, np.inf)

    # ---------
    # time loop

    # Beware of element-wise operations vs matrix operations !!!!
    nbsi = nbs
    bi = bounds
    indbsi = np.ones((nbsi,), dtype=bool)
    for ii in range(0, nt):

        if verb >= 1:
            msg = f"\ttime step {ii+1} / {nt} "
            print(msg, end='', flush=True)

        # update terms
        Tni, TTni, Tyni, yni, nchani = _update_TTyn(
            sparse=sparse,
            data_n=data_n,
            sigma=sigma,
            matrix=matrix,
            Tn=Tn,
            TTn=TTn,
            Tyn=Tyn,
            indok=indok,
            ii=ii,
            m3d=m3d,
            regul=regul,
        )

        if dcon is not None:
            ic = 0 if ii == 0 or not dcon['hastime'] else ii
            nbsi, indbsi, Tni, TTni, Tyni, yni, bi = _update_ttyn_constraints(
                sparse=sparse,
                Tni=Tni,
                TTni=TTni,
                Tyni=Tyni,
                yni=yni,
                bounds=bounds,
                ii=ic,
                dcon=dcon,
                regul=regul,
            )

        # solving
        (
            sol[ii, indbsi], mu[ii], chi2n[ii], regularity[ii],
            niter[ii], spec[ii],
        ) = dalgo['func'](
            Tn=Tni,
            TTn=TTni,
            Tyn=Tyni,
            R=R,
            yn=yni,
            # initial guess
            sol0=sol0[indbsi],
            mu0=mu0,
            # problem size
            nchan=nchani,
            nbs=nbsi,
            # parameters
            conv_crit=conv_crit,
            precond=precond,
            verb=verb,
            verb2head=verb2head,
            # quad-only
            func_val=func_val,
            func_jac=func_jac,
            func_hess=func_hess,
            bounds=bi,
            method=method,
            options=options,
            **kwdargs,
        )

        if dcon is not None:
            if dcon['coefs'] is not None:
                sol[ii, :] = (
                    dcon['coefs'][ic].dot(sol[ii, indbsi])
                    + dcon['offset'][ic, :]
                )
            else:
                sol[ii, :] = sol[ii, :] + dcon['offset'][ic, :]

        # post
        if chain:
            sol0[:] = sol[ii, :]
        mu0 = mu[ii]

        if verb == 1:
            msg = f"   chi2n = {chi2n[ii]:.3e}    niter = {niter[ii]}"
            print(msg, end='\n', flush=True)


# ##################################################################
# ##################################################################
#                   _compute time loop - TOMOTOK
# ##################################################################


def _compute_inv_loop_tomotok(
    # inputs
    dalgo=None,
    sol0=None,
    mu0=None,
    matrix=None,
    Tn=None,        # normalized geometry matrix (T)
    TTn=None,       # normalized tTT
    Tyn=None,       # normalized tTy
    R=None,
    precond=None,
    data_n=None,
    sigma=None,
    indok=None,
    # parameters
    m3d=None,
    conv_crit=None,
    isotropic=None,
    sparse=None,
    positive=None,
    chain=None,
    verb=None,
    kwdargs=None,
    method=None,
    options=None,
    dcon=None,
    regul=None,
    # output
    sol=None,
    mu=None,
    chi2n=None,
    regularity=None,
    niter=None,
    spec=None,
):

    # -----------------------------------
    # Getting initial solution - step 1/2

    nt, nchan = data_n.shape
    if not isinstance(R, np.ndarray):
        nbs = R[0].shape[0]
        R = (R,)
    else:
        nbs = R.shape[0]

    if dcon is not None:
        msg = "constraints not handled by tomotok algorithms"
        raise Exception(msg)

    if verb >= 2:
        form = "nchan * chi2n   +   mu *  R           "
        verb2head = f"\n\t\titer    {form} \t\t\t  tau  \t      conv"
    else:
        verb2head = None

    # ---------
    # time loop

    # Beware of element-wise operations vs matrix operations !!!!
    for ii in range(0, nt):

        if verb >= 1:
            msg = f"\ttime step {ii+1} / {nt} "
            print(msg, end='\n', flush=True)

        # update intermediates if multiple sigmas
        if sigma.shape[0] > 1:
            _update_TTyn(
                sparse=sparse,
                sigma=sigma,
                mati=mi,
                Tn=Tn,
                TTn=TTn,
                ii=ii,
            )

        Tyn[:] = Tn.T.dot(data_n[ii, :])

        # update terms
        Tni, TTni, Tyni, yni, nchani = _update_TTyn(
            sparse=sparse,
            data_n=data_n,
            sigma=sigma,
            matrix=matrix,
            Tn=Tn,
            TTn=TTn,
            Tyn=Tyn,
            indok=indok,
            ii=ii,
            m3d=m3d,
            regul=regul,
        )

        # solving
        (
            sol[ii, :], chi2n[ii], regularity[ii],
        ) = dalgo['func'](
            sig_norm=yni,
            gmat_norm=Tni,
            deriv=R,
            method=None,
            num=None,
            # additional
            nchan=nchan,
            **kwdargs,
        )

        # post
        if chain:
            sol0[:] = sol[ii, :]
        mu0 = mu[ii]

        if verb == 1:
            msg = f"   chi2n = {chi2n[ii]:.3e}    niter = {niter[ii]}"
            print(msg, end='\n', flush=True)


# ##################################################################
# ##################################################################
#                      utility
# ##################################################################


def _update_TTyn(
    sparse=None,
    data_n=None,
    sigma=None,
    matrix=None,
    Tn=None,
    TTn=None,
    Tyn=None,
    indok=None,
    ii=None,
    m3d=None,
    regul=None,
):
    # update matrix
    if indok is None:
        if m3d:
            mi = matrix[ii, :, :]
        else:
            mi = matrix
    else:
        if m3d:
            mi = matrix[ii, indok[ii, :], :]
        else:
            mi = matrix[indok[ii, :], :]

    # update sig
    if sigma.shape[0] == 1:
        if indok is None:
            sig = sigma[0, :]
        else:
            sig = sigma[0, indok[ii, :]]
    else:
        if indok is None:
            sig = sigma[ii, :]
        else:
            sig = sigma[ii, indok[ii, :]]

    # update data_n
    if indok is None:
        yn = data_n[ii, :]
    else:
        yn = data_n[ii, indok[ii, :]]

    # intermediates
    if indok is None:
        if sparse:
            Tn.data = scpsp.diags(1./sig).dot(mi).data
            TTn.data = Tn.T.dot(Tn).data
        else:
            Tn[...] = mi / sig[:, None]
            TTn[...] = Tn.T.dot(Tn)

    else:
        if sparse:
            Tn = scpsp.diags(1./sig).dot(mi)
            TTn = Tn.T.dot(Tn)
        else:
            Tn = mi / sig[:, None]
            TTn = Tn.T.dot(Tn)

    # Tyn (for reguarized algorithms)
    if regul:
        if indok is None:
            Tyn[:] = Tn.T.dot(yn)
        else:
            Tyn = Tn.T.dot(yn)

    return Tn, TTn, Tyn, yn, yn.size


def _update_ttyn_constraints(
    sparse=None,
    Tni=None,
    TTni=None,
    Tyni=None,
    yni=None,
    bounds=None,
    ii=None,
    dcon=None,
    regul=None,
):

    # regul => Tyni, TTni
    if regul:
        raise NotImplementedError()

    # yni
    yni = yni - Tni.dot(dcon['offset'][ii, :])

    # Tni
    Tni = Tni.dot(dcon['coefs'][ii])

    # indbsi
    indbsi = dcon['indbs'][ii]

    # nbsi
    nbsi = Tni.shape[1]

    # bounds
    if np.isscalar(bounds[0]):
        pass
    else:
        bounds = (bounds[0][indbsi], bounds[1][indbsi])

    return nbsi, indbsi, Tni, TTni, Tyni, yni, bounds


# ##################################################################
# ##################################################################
#               retrofit                   
# ##################################################################


def compute_retrofit_data(
    # resources
    coll=None,
    # inputs
    key=None,
    key_diag=None,
    key_matrix=None,
    key_profile2d=None,
    t=None,
    # parameters
    store=None,
    returnas=None,
):

    # ------------
    # check inputs
    # --------------

    (
        key, key_diag, key_cam, keybs, keym, mtype,
        key_matrix, key_profile2d,
        is2d, matrix, dindmat,
        hastime, t, keyt, reft, ref,
        nt, nchan, nbs,
        ist_mat, ist_prof, dind,
        store, returnas,
    ) = _compute_retrofit_data_check(
        # resources
        coll=coll,
        # inputs
        key=key,
        key_diag=key_diag,
        key_matrix=key_matrix,
        key_profile2d=key_profile2d,
        t=t,
        # parameters
        store=store,
        returnas=returnas,
    )

    # --------
    # prepare
    # --------------

    kmat = coll.dobj['geom matrix'][key_matrix]['data']
    gunits = coll.ddata[kmat[0]]['units']

    coefs = coll.ddata[key_profile2d]['data']

    # coefs
    if mtype == 'rect':
        indbs_tf = coll.select_bsplines(
            key=keybs,
            returnas='ind',
        )
        if hastime and ist_prof:
            coefs = coefs[:, indbs_tf[0], indbs_tf[1]]
        else:
            coefs = coefs[indbs_tf[0], indbs_tf[1]]

    # --------
    # compute
    # --------------

    # time-dependent
    if hastime:

        # retro = np.full((nt, nchan, nbs), np.nan)

        # get time indices
        if ist_mat:
            if dind.get(kmat[0], {}).get('ind') is not None:
                imat = dind[kmat[0]]['ind']
            else:
                imat = np.arange(nt)

        if ist_prof:
            if dind.get(key_profile2d, {}).get('ind') is not None:
                iprof = dind[key_profile2d]['ind']
            else:
                iprof = np.arange(nt)

        # compute matrix product
        if ist_mat and ist_prof:
            retro = np.array([
                matrix[imat[ii], :, :].dot(coefs[iprof[ii], :])
                for ii in range(nt)
            ])
        elif ist_mat:
            retro = np.array([
                matrix[imar[ii], :, :].dot(coefs)
                for ii in range(nt)
            ])
        elif ist_prof:
            retro = np.array([
                matrix.dot(coefs[iprof[ii], :])
                for ii in range(nt)
            ])
    else:
        retro = matrix.dot(coefs)

    # --------------
    # format
    # --------------

    i0 = 0
    dout = {}
    for ii, k0 in enumerate(key_cam):
        npix = coll.dobj['camera'][k0]['dgeom']['pix_nb']
        ind = i0 + np.arange(0, npix)

        # extract relevant part
        retroi = retro[:, ind] if hastime else retro[ind]
        refi = list(ref)
        axis = refi.index(None)

        # reshape if 2d
        if is2d:
            sh = list(retroi.shape)
            sh[axis] = coll.dobj['camera'][k0]['dgeom']['shape']
            sh = tuple(np.r_[sh[:axis], sh[axis], sh[axis+1:]].astype(int))
            retroi = retroi.reshape(sh)

        # ref
        refi[axis] = coll.dobj['camera'][k0]['dgeom']['ref']
        refi = tuple(np.r_[refi[:axis], refi[axis], refi[axis+1:]])

        # dict
        dout[k0] = {
            'data': retroi,
            'ref': refi,
        }
        i0 += npix

    units = coll.ddata[key_profile2d]['units'] * gunits

    # --------------
    # store
    # --------------

    if store:
        _class8_compute_signal._store(
            coll=coll,
            key=key,
            key_diag=key_diag,
            dout=dout,
            units=units,
            key_matrix=key_matrix,
        )

    # -------------
    # return 
    # --------------

    if returnas is dict:
        return dout


# ###################
#   checking
# ###################


def _compute_retrofit_data_check(
    # resources
    coll=None,
    # inputs
    key=None,
    key_diag=None,
    key_matrix=None,
    key_profile2d=None,
    t=None,
    # parameters
    store=None,
    returnas=None,
):

    #----------
    # keys

    # key_diag
    lok = list(coll.dobj.get('diagnostic', {}).keys())
    key_diag = ds._generic_check._check_var(
        key_diag, 'key_diag',
        types=str,
        allowed=lok,
    )
    is2d = coll.dobj['diagnostic'][key_diag]['is2d']

    # key
    dsig = coll.dobj['diagnostic'][key_diag].get('dsignal', {})
    lout = list(dsig.keys())
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        excluded=lout,
    )

    # key_matrix
    lok = coll.dobj.get('geom matrix', {}).keys()
    key_matrix = ds._generic_check._check_var(
        key_matrix, 'key_matrix',
        types=str,
        allowed=lok,
    )

    key_cam = coll.dobj['geom matrix'][key_matrix]['camera']
    keybs = coll.dobj['geom matrix'][key_matrix]['bsplines']
    keym = coll.dobj['bsplines'][keybs]['mesh']
    mtype = coll.dobj[coll._which_mesh][keym]['type']

    matrix, ref, dindmat = coll.get_geometry_matrix_concatenated(key=key_matrix)
    nchan, nbs = matrix.shape[-2:]
    refbs = ref[-1]

    # key_pofile2d
    lok = [
        k0 for k0, v0 in coll.ddata.items()
        if v0['bsplines'] == keybs
    ]
    key_profile2d = ds._generic_check._check_var(
        key_profile2d, 'key_profile2d',
        types=str,
        allowed=lok,
    )

    # time management
    lkmat = coll.dobj['geom matrix'][key_matrix]['data']
    hastime, reft, keyt, t_out, dind = coll.get_time_common(
        keys=lkmat + [key_profile2d],
        t=t,
        ind_strict=False,
    )
    if hastime and t_out is not None and reft is None:
        reft = f'{key}-nt'
        keyt = f'{key}-t'

    ist_mat = coll.get_time(key=lkmat[0])[0]
    ist_prof = coll.get_time(key=key_profile2d)[0]

    # reft, keyt and refs
    if hastime and t_out is not None:
        nt = t_out.size
        ref = (reft, None)
    else:
        nt = 0
        reft = None
        keyt = None
        ref = (None,)

    # store
    store = ds._generic_check._check_var(
        store, 'store',
        types=bool,
        default=True,
    )

    # returnas
    returnas = ds._generic_check._check_var(
        returnas, 'returnas',
        default=False if store else dict,
        allowed=[False, dict],
    )


    return (
        key, key_diag, key_cam, keybs, keym, mtype,
        key_matrix, key_profile2d,
        is2d, matrix, dindmat,
        hastime, t_out, keyt, reft, ref,
        nt, nchan, nbs,
        ist_mat, ist_prof, dind,
        store, returnas,
    )



# -*- coding: utf-8 -*-


# Built-in
import copy


# Common
import numpy as np
import datastock as ds


_DPINHOLE = {
    # center
    'x': None,
    'y': None,
    'R': 3.,
    'z': 0.,
    'phi': 0.,
    # angles
    'theta': np.pi,
    'dphi': 0.,
    'tilt': 0,
}


# ##########################################################################
# ##########################################################################
#                   Add pinhole camera
# ##########################################################################


def add_camera_pinhole(
    coll=None,
    key=None,
    key_pinhole=None,
    key_diag=None,
    cam_type=None,
    # position
    x=None,
    y=None,
    R=None,
    z=None,
    phi=None,
    # orientation
    theta=None,
    dphi=None,
    tilt=None,
    # camera
    focal=None,
    pix_nb=None,
    pix_size=None,
    pix_spacing=None,
    # pinhole
    pinhole_radius=None,
    pinhole_size=None,
    # reflections
    reflections_nb=None,
    reflections_type=None,
    # diagnostic
    compute=None,
    config=None,
    length=None,
):

    # --------------
    # check inputs

    key, key_pinhole, key_diag, newdiag = _check_camera_pinhole(
        coll=coll,
        key=key,
        key_pinhole=key_pinhole,
        key_diag=key_diag,
    )

    # ------------------------
    # compute pinhole position

    pc, nin, e0, e1 = _pinhole_position(
        # center
        x=x,
        y=y,
        R=R,
        z=z,
        phi=phi,
        # angles
        theta=theta,
        dphi=dphi,
        tilt=tilt,
    )

    # ------------------------
    # compute pinhole contour

    pin_out0, pin_out1 = _pinhole_contour(
        pinhole_radius=pinhole_radius,
        pinhole_size=pinhole_size,
    )

    dgeom_pin = {
        'cent': pc,
        'nin': nin,
        'e0': e0,
        'e1': e1,
        'outline_x0': pin_out0,
        'outline_x1': pin_out1,
    }

    # ------------------------
    # compute camera position

    dgeom_cam, cam_type = _camera_position(
        dgeom_pin=dgeom_pin,
        cam_type=cam_type,
        focal=focal,
        pix_nb=pix_nb,
        pix_size=pix_size,
        pix_spacing=pix_spacing,
    )

    # -------
    # store

    _add_camera_pinhole_store(
        coll=coll,
        key=key,
        key_pinhole=key_pinhole,
        key_diag=key_diag,
        cam_type=cam_type,
        dgeom_pin=dgeom_pin,
        dgeom_cam=dgeom_cam,
        newdiag=newdiag,
        # reflections
        reflections_nb=reflections_nb,
        reflections_type=reflections_type,
        compute=compute,
        config=config,
        length=length,
    )


def _check_camera_pinhole(
    coll=None,
    key=None,
    key_pinhole=None,
    key_diag=None,
):
    # key
    lout = list(coll.dobj.get('camera', {}).keys())
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        excluded=lout,
    )

    # key_pinhole
    if key_pinhole is None:
        key_pinhole = f'{key}_pin'

    lout = list(coll.dobj.get('aperture', {}).keys())
    key_pinhole = ds._generic_check._check_var(
        key_pinhole, 'key_pinhole',
        types=str,
        excluded=lout,
    )

    # key_diag
    if key_diag is not None:
        lok = list(coll.dobj.get('diagnostic', {}).keys())
        key_diag = ds._generic_check._check_var(
            key_diag, 'key_diag',
            types=str,
        )

        newdiag = key_diag not in lok
    else:
        newdiag = True

    return key, key_pinhole, key_diag, newdiag


def _pinhole_position(
    # center
    x=None,
    y=None,
    R=None,
    z=None,
    phi=None,
    # angles
    theta=None,
    dphi=None,
    tilt=None,
    # default
    ddef=None,
):

    # -------------
    # check inputs

    # ddef
    if ddef is None:
        ddef = _DPINHOLE

    # (x, y) vs (R, phi)
    lc = [
        x is not None or y is not None,
        R is not None or phi is not None,
    ]
    if np.sum(lc) > 1:
        msg = "Please provide (x, y) xor (R, phi) !"
        raise Exception(msg)

    # (x, y) vs (R, phi)
    if lc[0]:
        if x is None:
            x = ddef['x']
        if y is None:
            y = ddef['y']
        R = np.hypot(x, y)
        phi = np.arctan2(y, x)
    else:
        if R is None:
            R = ddef['R']
        if phi is None:
            phi = ddef['phi']
        x = R * np.cos(phi)
        y = R * np.sin(phi)

    # z
    if z is None:
        z = ddef['z']

    # dphi
    if dphi is None:
        dphi = ddef['dphi']

    # theta
    if theta is None:
        theta = ddef['theta']

    # tilt
    if tilt is None:
        tilt = ddef['tilt']

    # -------------
    # compute

    # unit vectors of reference
    eR = np.r_[np.cos(phi), np.sin(phi), 0.]
    ephi = np.r_[-np.sin(phi), np.cos(phi), 0.]
    er = np.cos(theta) * eR + np.sin(theta) * np.r_[0, 0, 1]
    etheta = -np.sin(theta) * eR + np.cos(theta) * np.r_[0, 0, 1]
    e0bis = -np.cos(dphi) * ephi + np.sin(dphi) * er

    # translation
    pc = np.r_[x, y, z]

    # unit vectors after rotation
    nin = np.cos(dphi) * er + np.sin(dphi) * ephi
    e0 = np.cos(tilt) * e0bis + np.sin(tilt) * etheta
    e1 = np.cross(nin, e0)

    # safety check
    nin, e0, e1 = ds._generic_check._check_vectbasis(
        e0=nin,
        e1=e0,
        e2=e1,
        dim=3,
        tol=1e-12,
    )

    return pc, nin, e0, e1


def _pinhole_contour(
    pinhole_radius=None,
    pinhole_size=None,
):

    # ------
    # check

    if pinhole_radius is not None and pinhole_size is not None:
        msg = (
            "Please provide pinhole_radius xor pinhole_size!\n"
            "\t- pinhole_radius: created a circular pinhole\n"
            "\t- pinhole_size: created a rectangular pinhole\n"
        )
        raise Exception(msg)

    # ----------
    # compute

    # circular
    if pinhole_radius is not None:

        # check
        pinhole_radius = ds._generic_check._check_var(
            pinhole_radius, 'pinhole_radius',
            types=(int, float),
            sign='> 0.',
        )

        # compute
        theta = np.pi * np.linspace(-1, 1, 50)[:-1]
        out0 = pinhole_radius * np.cos(theta)
        out1 = pinhole_radius * np.sin(theta)

    else:
        # check
        if np.isscalar(pinhole_size):
            pinhole_size = [pinhole_size, pinhole_size]

        pinhole_size = ds._generic_check._check_flat1darray(
            pinhole_size, 'pinhole_size',
            dtype=float,
            size=2,
            sign='> 0.',
            norm=False,
            can_be_None=False,
        )

        # compute
        out0 = pinhole_size[0] * np.r_[-1, 1, 1, -1]
        out1 = pinhole_size[1] * np.r_[-1, -1, 1, 1]

    return out0, out1

def _camera_position(
    dgeom_pin=None,
    cam_type=None,
    focal=None,
    pix_nb=None,
    pix_size=None,
    pix_spacing=None,
):

    # -------------
    # check

    # cam_type
    cam_type = ds._generic_check._check_var(
        cam_type, 'cam_type',
        types=str,
        default='1d',
        allowed=['1d', '2d'],
    )

    # cam_type
    focal = ds._generic_check._check_var(
        focal,'focal',
        types=(int, float),
        default=0.1,
        sign='>0.',
    )

    if cam_type == '1d':

        # pix_nb
        pix_nb = ds._generic_check._check_var(
            pix_nb, 'pix_nb',
            types=int,
            sign='> 0',
        )

        # pix_spacing
        pix_spacing = ds._generic_check._check_var(
            pix_spacing, 'pix_spacing',
            types=(float, int),
            default=0,
            sign='>= 0',
        )

    else:

        # pix_nb
        if np.isscalar(pix_nb):
            pix_nb = pix_nb * np.r_[1, 1]

        pix_nb = ds._generic_check._check_flat1darray(
            pix_nb, 'pix_nb',
            size=2,
            dtype=int,
            sign='> 0',
        )

        # pix_spacing
        if pix_spacing is None:
            pix_spacing = 0.
        if np.isscalar(pix_spacing):
            pix_spacing = pix_spacing * np.r_[1, 1]

        pix_spacing = ds._generic_check._check_flat1darray(
            pix_spacing, 'pix_spacing',
            size=2,
            dtype=float,
            sign='>= 0',
        )

    # pix_size
    if np.isscalar(pix_size):
        pix_size = pix_size * np.r_[1, 1]

    pix_size = ds._generic_check._check_flat1darray(
        pix_size, 'pix_size',
        size=2,
        dtype=float,
        sign='> 0',
    )

    # ----------
    # compute

    dgeom_cam = copy.deepcopy(dgeom_pin)
    cent = dgeom_pin['cent'] - focal * dgeom_pin['nin']

    if cam_type == '1d':

        # vectors
        for k0 in ['nin', 'e0', 'e1']:
            for ii, ss in enumerate(['x', 'y', 'z']):
                dgeom_cam[f'{k0}_{ss}'] = dgeom_cam[k0][ii]
            del dgeom_cam[k0]

        # cents
        dd = (pix_size[0] + pix_spacing) * np.arange(0, pix_nb)
        dd = dd - np.mean(dd)
        dgeom_cam['cents_x'] = cent[0] + dd * dgeom_pin['e0'][0]
        dgeom_cam['cents_y'] = cent[1] + dd * dgeom_pin['e0'][1]
        dgeom_cam['cents_z'] = cent[2] + dd * dgeom_pin['e0'][2]
        del dgeom_cam['cent']

    else:
        for ii in [0, 1]:
            dd = (pix_size[ii] + pix_spacing[ii]) * np.arange(0, pix_nb[ii])
            dgeom_cam[f'cents_x{ii}'] = dd - np.mean(dd)

        dgeom_cam['cent'] = cent

    # -------------
    # pixel outline

    dgeom_cam['outline_x0'] = 0.5 * pix_size[0] * np.r_[-1, 1, 1, -1]
    dgeom_cam['outline_x1'] = 0.5 * pix_size[1] * np.r_[-1, -1, 1, 1]

    return dgeom_cam, cam_type


def _add_camera_pinhole_store(
    coll=None,
    key=None,
    key_pinhole=None,
    key_diag=None,
    cam_type=None,
    dgeom_pin=None,
    dgeom_cam=None,
    newdiag=None,
    # reflections
    reflections_nb=None,
    reflections_type=None,
    compute=None,
    config=None,
    length=None,
):

    # compute
    compute = ds._generic_check._check_var(
        compute, 'compute',
        types=bool,
        default=True,
    )

    # pinhole aperture
    coll.add_aperture(
        key=key_pinhole,
        **dgeom_pin,
    )

    # camera
    if cam_type == '1d':
        coll.add_camera_1d(
            key=key,
            dgeom=dgeom_cam,
            dmat=None,
        )
    else:
        coll.add_camera_2d(
            key=key,
            dgeom=dgeom_cam,
            dmat=None,
        )

    # doptics
    doptics = {key: [key_pinhole]}

    # diagnostic
    if newdiag is False:

        # delete and recreate diagnostic
        dop = {
            k0: v0['optics']
            for k0, v0 in coll._dobj['diagnostic'][key_diag]['doptics'].items()
        }
        dop.update(doptics)
        doptics = dop
        coll.remove_diagnostic(key=key_diag)

    # create
    coll.add_diagnostic(
        key=key_diag,
        doptics=doptics,
        reflections_nb=reflections_nb,
        reflections_type=reflections_type,
        compute=compute,
        config=config,
        length=length,
    )
